from dataclasses import dataclass


@dataclass
class Game:
    """
    Represents a Twitch category

    Attributes:
        game_id (int): An ID that identifies the category or game
        name (str): The category’s or game’s name
        box_art_url (str | None): A URL to the category’s or game’s box art
        igdb_id (str | None): The ID that IGDB uses to identify this game
    """

    game_id: str
    name: str
    box_art_url: str | None = None
    igdb_id: str | None = None



from dataclasses import dataclass

from ..dataclasses import Channel, User


@dataclass
class Subscription:
    """
    Represents a subscription

    Attributes:
        channel (Channel): The channel
        gifter (User): The user that gifted the subscription to the user
        is_gift (bool): A Boolean value that determines whether the subscription is a gift subscription
        tier (str): The type of subscription
            Possible values: 1000, 2000, 3000
        plan_name (str | None): The name of the subscription
        user (User | None): The subscribing user
    """

    channel: Channel
    gifter: User
    is_gift: bool
    tier: str
    plan_name: str | None = None
    user: User | None = None



from dataclasses import dataclass
from datetime import datetime


@dataclass
class DropEntitlement:
    """
    Represents a entitlement

    Attributes:
        drop_entitlement_id (str): An ID that identifies the entitlement
        benefit_id (str): An ID that identifies the benefit (reward)
        timestamp (datetime): The UTC date and time (in RFC3339 format) of when the entitlement was granted
        user_id (str): An ID that identifies the user who was granted the entitlement
        game_id (str): An ID that identifies the game the user was playing when the reward was entitled
        fulfillment_status (str): The entitlement’s fulfillment status
            Possible values: CLAIMED, FULFILLED
        last_updated (datetime): The UTC date and time (in RFC3339 format) of when the entitlement was last updated
    """

    drop_entitlement_id: str
    benefit_id: str
    timestamp: datetime
    user_id: str
    game_id: str
    fulfillment_status: str
    last_updated: datetime



from dataclasses import dataclass

from ..dataclasses import ProductCost


@dataclass
class Product:
    """
    Represents a digital product

    Attributes:
        sku (str): An ID that identifies the digital product
        cost (ProductCost): The digital product’s cost
        in_development (bool): A Boolean value that determines whether the product is in development
        display_name (str): The name of the digital product
        expiration (str): This field is always empty since you may purchase only unexpired products
        broadcast (bool): A Boolean value that determines whether the data was broadcast to all instances of the extension
        domain (str | None): The product's domain
    """

    sku: str
    cost: ProductCost
    in_development: bool
    display_name: str
    expiration: str
    broadcast: bool
    domain: str | None = None



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import CheermoteTier


@dataclass
class Cheermote:
    """
    Represents a Cheermote

    Attributes:
        prefix (str): The name portion of the Cheermote string that you use in chat to cheer Bits
            The full Cheermote string is the concatenation of {prefix} + {number of Bits}
        tiers (list[CheermoteTier]): A list of tier levels that the Cheermote supports
            Each tier identifies the range of Bits that you can cheer at that tier level and an image that graphically identifies the tier level
        type (str): The type of Cheermote
            Possible values: global_first_party, global_third_party, channel_custom, display_only, sponsored
        order (int): The order that the Cheermotes are shown in the Bits card
        last_updated (datetime): The date and time, in RFC3339 format, when this Cheermote was last updated
        is_charitable (bool): A Boolean value that indicates whether this Cheermote provides a charitable contribution match during charity campaigns
    """

    prefix: str
    tiers: list[CheermoteTier]
    type: str
    order: int
    last_updated: datetime
    is_charitable: bool



from dataclasses import dataclass

from ..dataclasses import BadgeVersion


@dataclass
class Badge:
    """
    Represents a chat badge

    Attributes:
        set_id (str): ID for the chat badge set
        versions (list[BadgeVersion]): Chat badge versions for the set
    """

    set_id: str
    versions: list[BadgeVersion]



from dataclasses import dataclass


@dataclass
class ProductCost:
    """
    Represents a digital product's cost

    Attributes:
        amount (int): The amount exchanged for the digital product
        type (str): The type of currency exchanged
            Possible values: bits
    """

    amount: int
    type: str



from dataclasses import dataclass

from ..dataclasses import User


@dataclass
class Predictor:
    """
    Represents a predictor in a prediction

    Attributes:
        user (User): The viewer
        channel_points_used (int): The number of Channel Points the viewer spent
        channel_points_won (int): The number of Channel Points distributed to the viewer
    """

    user: User
    channel_points_used: int
    channel_points_won: int



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Game


@dataclass
class StreamScheduleSegment:
    """
    Represents a scheduled broadcast in a channel's streaming schedule

    Attributes:
        segment_id (str): An ID that identifies this broadcast segment
        start_time (datetime): The UTC date and time (in RFC3339 format) of when the broadcast starts
        end_time (datetime): The UTC date and time (in RFC3339 format) of when the broadcast ends
        title (str): The broadcast segment’s title
        canceled_until (datetime): Indicates whether the broadcaster canceled this segment of a recurring broadcast
        category (Game): The type of content that the broadcaster plans to stream
        is_recurring (bool): A Boolean value that determines whether the broadcast is part of a recurring series that streams at the same time each week or is a one-time broadcast
    """

    segment_id: str
    start_time: datetime
    end_time: datetime
    title: str
    canceled_until: datetime
    category: Game
    is_recurring: bool



class TwitchPyBException(Exception):
    """
    Base exception class for TwitchPy
    """

    def __init__(self, message):
        super().__init__(message)


class AppTokenError(TwitchPyBException):
    """
    Exception raised when an error trying to get an app token happens
    """


class UserTokenError(TwitchPyBException):
    """
    Exception raised when a function that requires an user token is called without an user token
    """


class ClientError(TwitchPyBException):
    """
    Exception raised when a bad request is made by a client
    """



import socket
import ssl
from typing import Callable

from .client import Client
from .dataclasses import Message

_IRC_SERVER = "irc.chat.twitch.tv"
_IRC_PORT = 6697

_DEFAULT_TIMEOUT = 10


class Bot:
    """
    Represents a bot
    """

    def __init__(
        self,
        oauth_token: str,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        tokens_path: str,
        username: str,
        channels: list[str],
        command_prefix: str,
        authorization_code: str | None = None,
        jwt_token: str | None = None,
        ready_message: str | None = None,
    ):
        """
        Args:
            oauth_token (str): OAuth token
            client_id (str): Client ID
            client_secret (str): Client secret
            redirect_uri (str): Redirect URI
            tokens_path (str): Path of tokens file (file included)
            username (str): Name of the bot
            channels (list): Names of channels the bot will access
            command_prefix (str): Prefix of the commands the bot will recognize
            code (str): Authorization code for getting an user token
            jwt_token (str): JWT Token
            ready_message (str): Message that the bot will send through the chats of the channels it access
        """

        self.client = Client(
            client_id,
            client_secret,
            redirect_uri,
            tokens_path,
            authorization_code,
            jwt_token,
        )
        self.__oauth_token = oauth_token
        self.__finish = False
        self.username = username

        self.channels = []

        for channel in channels:
            self.channels.append(channel.replace("@", "").lower())

        self.command_prefix = command_prefix
        self.ready_message = ready_message if ready_message is not None else ""

        self.custom_methods_before_join_channel = {}
        self.methods_before_join_channel_to_remove = []
        self.custom_methods_after_join_channel = {}
        self.methods_after_join_channel_to_remove = []

        self.custom_methods_before_leave_channel = {}
        self.methods_before_leave_channel_to_remove = []
        self.custom_methods_after_leave_channel = {}
        self.methods_after_leave_channel_to_remove = []

        self.custom_checks = {}
        self.checks_to_remove = []
        self.custom_listeners = {}
        self.listeners_to_remove = []
        self.custom_commands = {}
        self.commands_to_remove = []
        self.custom_methods_before_commands = {}
        self.methods_before_commands_to_remove = []
        self.custom_methods_after_commands = {}
        self.methods_after_commands_to_remove = []

        self.custom_methods_after_clearchat = {}
        self.methods_after_clearchat_to_remove = []

        self.custom_methods_after_delete_message = {}
        self.methods_after_delete_message_to_remove = []

        self.custom_methods_after_bot_connected = {}
        self.methods_after_bot_connected_to_remove = []

        self.custom_methods_after_toggle_host = {}
        self.methods_after_toggle_host_to_remove = []

        self.custom_methods_after_server_reconnect = {}
        self.methods_after_server_reconnect_to_remove = []

        self.custom_methods_after_channel_change = {}
        self.methods_after_channel_change_to_remove = []

        self.custom_methods_after_event = {}
        self.methods_after_event_to_remove = []

        self.custom_methods_after_user_join = {}
        self.methods_after_user_join_to_remove = []

        self.custom_methods_after_whisper = {}
        self.methods_after_whisper_to_remove = []

        self.irc = ssl.SSLContext().wrap_socket(socket.socket())

    def __send_command(self, command: str, args: str) -> None:
        print(f"{command} < {args}")

        self.irc.send((f"{command} {args}" + "\r\n").encode())

    def __send_join(self, channel: str) -> None:
        self.__send_command("JOIN", f"#{channel}")

    def __send_nick(self, username: str) -> None:
        self.__send_command("NICK", username)

    def __send_part(self, channel: str) -> None:
        self.__send_command("PART", f"#{channel}")

    def __send_pass(self, oauth_token: str) -> None:
        self.__send_command("PASS", oauth_token)

    def __send_pong(self, text: str) -> None:
        self.__send_command("PONG", f":{text}")

    def __send_privmsg(self, channel: str, text: str) -> None:
        self.__send_command("PRIVMSG", f"#{channel} :{text}")

    def __login(self) -> None:
        self.__send_pass(self.__oauth_token)
        self.__send_nick(self.username)

    def __request_irc_capabilities(self) -> None:
        self.__send_command("CAP REQ", ":twitch.tv/commands twitch.tv/membership")

    def join_channel(self, channel: str) -> None:
        """
        Makes the bot to join into a channel

        Args:
            channel (str): The channel to join
        """

        self.__execute_methods_before_join_channel(channel)

        self.__send_join(channel)
        self.__send_privmsg(channel, self.ready_message)

        self.__execute_methods_after_join_channel(channel)

        self.__remove_methods_before_join_channel()
        self.__remove_methods_after_join_channel()

    def leave_channel(self, channel: str) -> None:
        """
        Makes the bot to leave into a channel

        Args:
            channel (str): The channel to leave
        """

        self.__execute_methods_before_leave_channel(channel)

        self.__send_part(channel)

        self.__execute_methods_after_leave_channel(channel)

        self.__remove_methods_before_leave_channel()
        self.__remove_methods_after_leave_channel()

    def __connect(self) -> None:
        self.irc.settimeout(_DEFAULT_TIMEOUT)
        self.irc.connect((_IRC_SERVER, _IRC_PORT))

        self.__login()
        self.__request_irc_capabilities()

        for channel in self.channels:
            self.join_channel(channel)

    def run(self) -> None:
        """
        Runs the bot
        """

        self.__connect()
        self.__loop()

    def stop(self) -> None:
        """
        Stops the bot
        """

        self.__finish = True

    def __get_user_from_prefix(self, prefix: str) -> str | None:
        domain = prefix.split("!")[0]

        if domain.endswith(".tmi.twitch.tv"):
            return domain.replace(".tmi.twitch.tv", "")

        if "tmi.twitch.tv" not in domain:
            return domain

        return None

    def __remove_prefix(self, string: str, prefix: str) -> str:
        if not string.startswith(prefix):
            return string

        return string[len(prefix) :]

    def __parse_message(self, received_msg: str) -> Message:
        parts = received_msg.split(" ")

        prefix = None
        user = None
        channel = None
        irc_command = None
        irc_args = None
        text = None
        text_command = None
        text_args = None

        if parts[0].startswith(":"):
            prefix = self.__remove_prefix(parts[0], ":")
            user = self.__get_user_from_prefix(prefix)
            parts = parts[1:]

        text_start = next(
            (idx for idx, part in enumerate(parts) if part.startswith(":")), None
        )

        if text_start is not None:
            text_parts = parts[text_start:]
            text_parts[0] = text_parts[0][1:]
            text = " ".join(text_parts)

            if text_parts[0].startswith(self.command_prefix):
                text_command = self.__remove_prefix(text_parts[0], self.command_prefix)
                text_args = text_parts[1:]

            parts = parts[:text_start]

        irc_command = parts[0]
        irc_args = parts[1:]

        hash_start = next(
            (idx for idx, part in enumerate(irc_args) if part.startswith("#")), None
        )

        if hash_start is not None:
            channel = irc_args[hash_start][1:]

        message = Message(
            prefix,
            user,
            channel,
            irc_command,
            irc_args,
            text,
            text_command,
            text_args,
        )

        return message

    def __execute_methods_before_join_channel(self, channel: str) -> None:
        for method in self.custom_methods_before_join_channel.values():
            method(channel)

    def __remove_methods_before_join_channel(self) -> None:
        for method in self.methods_before_join_channel_to_remove:
            if method in self.custom_methods_before_join_channel:
                self.custom_methods_before_join_channel.pop(method)

        self.methods_before_join_channel_to_remove = []

    def __execute_methods_after_join_channel(self, channel: str) -> None:
        for method in self.custom_methods_after_join_channel.values():
            method(channel)

    def __remove_methods_after_join_channel(self) -> None:
        for method in self.methods_after_join_channel_to_remove:
            if method in self.custom_methods_after_join_channel:
                self.custom_methods_after_join_channel.pop(method)

        self.methods_after_join_channel_to_remove = []

    def __execute_methods_before_leave_channel(self, channel: str) -> None:
        for method in self.custom_methods_before_leave_channel.values():
            method(channel)

    def __remove_methods_before_leave_channel(self) -> None:
        for method in self.methods_before_leave_channel_to_remove:
            if method in self.custom_methods_before_leave_channel:
                self.custom_methods_before_leave_channel.pop(method)

        self.methods_before_leave_channel_to_remove = []

    def __execute_methods_after_leave_channel(self, channel: str) -> None:
        for method in self.custom_methods_after_leave_channel.values():
            method(channel)

    def __remove_methods_after_leave_channel(self) -> None:
        for method in self.methods_after_leave_channel_to_remove:
            if method in self.custom_methods_after_leave_channel:
                self.custom_methods_after_leave_channel.pop(method)

        self.methods_after_leave_channel_to_remove = []

    def __execute_checks(self) -> None:
        for check in self.custom_checks.values():
            check()

    def __remove_checks(self) -> None:
        for check in self.checks_to_remove:
            if check in self.custom_checks:
                self.custom_checks.pop(check)

        self.checks_to_remove = []

    def __execute_listeners(self, message: Message) -> None:
        for listener in self.custom_listeners.values():
            listener(message)

    def __remove_listeners(self) -> None:
        for listener in self.listeners_to_remove:
            if listener in self.custom_listeners:
                self.custom_listeners.pop(listener)

        self.listeners_to_remove = []

    def __execute_command(self, message: Message) -> None:
        self.custom_commands[message.text_command](message)

    def __remove_commands(self) -> None:
        for command in self.commands_to_remove:
            if command in self.custom_commands:
                self.custom_commands.pop(command)

        self.commands_to_remove = []

    def __execute_methods_before_commands(self, message: Message) -> None:
        for before in self.custom_methods_before_commands.values():
            before(message)

    def __remove_methods_before_commands(self) -> None:
        for method in self.methods_before_commands_to_remove:
            if method in self.custom_methods_before_commands:
                self.custom_methods_before_commands.pop(method)

        self.methods_before_commands_to_remove = []

    def __execute_methods_after_commands(self, message: Message) -> None:
        for after in self.custom_methods_after_commands.values():
            after(message)

    def __remove_methods_after_commands(self) -> None:
        for method in self.methods_after_commands_to_remove:
            if method in self.custom_methods_after_commands:
                self.custom_methods_after_commands.pop(method)

        self.methods_after_commands_to_remove = []

    def __execute_methods_after_clearchat(self, message: Message) -> None:
        for method in self.custom_methods_after_clearchat.values():
            method(message)

    def __remove_methods_after_clearchat(self) -> None:
        for method in self.methods_after_clearchat_to_remove:
            if method in self.custom_methods_after_clearchat:
                self.custom_methods_after_clearchat.pop(method)

        self.methods_after_clearchat_to_remove = []

    def __execute_methods_after_delete_message(self, message: Message) -> None:
        for method in self.custom_methods_after_delete_message.values():
            method(message)

    def __remove_methods_after_delete_message(self) -> None:
        for method in self.methods_after_delete_message_to_remove:
            if method in self.custom_methods_after_delete_message:
                self.custom_methods_after_delete_message.pop(method)

        self.methods_after_delete_message_to_remove = []

    def __execute_methods_after_bot_connected(self, message: Message) -> None:
        for method in self.custom_methods_after_bot_connected.values():
            method(message)

    def __remove_methods_after_bot_connected(self) -> None:
        for method in self.methods_after_bot_connected_to_remove:
            if method in self.custom_methods_after_bot_connected:
                self.custom_methods_after_bot_connected.pop(method)

        self.methods_after_bot_connected_to_remove = []

    def __execute_methods_after_toggle_host(self, message: Message) -> None:
        for method in self.custom_methods_after_toggle_host.values():
            method(message)

    def __remove_methods_after_toggle_host(self) -> None:
        for method in self.methods_after_toggle_host_to_remove:
            if method in self.custom_methods_after_toggle_host:
                self.custom_methods_after_toggle_host.pop(method)

        self.methods_after_toggle_host_to_remove = []

    def __execute_methods_after_server_reconnect(self, message: Message) -> None:
        for method in self.custom_methods_after_server_reconnect.values():
            method(message)

    def __remove_methods_after_server_reconnect(self) -> None:
        for method in self.methods_after_server_reconnect_to_remove:
            if method in self.custom_methods_after_server_reconnect:
                self.custom_methods_after_server_reconnect.pop(method)

        self.methods_after_server_reconnect_to_remove = []

    def __execute_methods_after_channel_change(self, message: Message) -> None:
        for method in self.custom_methods_after_channel_change.values():
            method(message)

    def __remove_methods_after_channel_change(self) -> None:
        for method in self.methods_after_channel_change_to_remove:
            if method in self.custom_methods_after_channel_change:
                self.custom_methods_after_channel_change.pop(method)

        self.methods_after_channel_change_to_remove = []

    def __execute_methods_after_event(self, message: Message) -> None:
        for method in self.custom_methods_after_event.values():
            method(message)

    def __remove_methods_after_event(self) -> None:
        for method in self.methods_after_event_to_remove:
            if method in self.custom_methods_after_event:
                self.custom_methods_after_event.pop(method)

        self.methods_after_event_to_remove = []

    def __execute_methods_after_user_join(self, message: Message) -> None:
        for method in self.custom_methods_after_user_join.values():
            method(message)

    def __remove_methods_after_user_join(self) -> None:
        for method in self.methods_after_user_join_to_remove:
            if method in self.custom_methods_after_user_join:
                self.custom_methods_after_user_join.pop(method)

        self.methods_after_user_join_to_remove = []

    def __execute_methods_after_whisper(self, message: Message) -> None:
        for method in self.custom_methods_after_whisper.values():
            method(message)

    def __remove_methods_after_whisper(self) -> None:
        for method in self.methods_after_whisper_to_remove:
            if method in self.custom_methods_after_whisper:
                self.custom_methods_after_whisper.pop(method)

        self.methods_after_whisper_to_remove = []

    def __handle_notice(self, message: Message) -> None:
        print(f"{message.irc_command} > [{message.channel}]: {message.text}")

    def __handle_part(self, message: Message) -> None:
        print(f"{message.irc_command} > [{message.channel}] {message.user}")

        self.__execute_methods_after_leave_channel(
            message.channel if message.channel is not None else ""
        )
        self.__remove_methods_after_leave_channel()

    def __handle_ping(self, message: Message) -> None:
        print(f"{message.irc_command} > :{message.text}")

        self.__send_pong(message.text if message.text is not None else "")

    def __handle_privmsg(self, message: Message) -> None:
        print(
            f"{message.irc_command} > [{message.channel}] {message.user}: {message.text}"
        )

        self.__execute_listeners(message)
        self.__remove_listeners()

        if message.text_command in self.custom_commands:
            self.__execute_methods_before_commands(message)
            self.__remove_methods_before_commands()
            self.__execute_command(message)
            self.__remove_commands()
            self.__execute_methods_after_commands(message)
            self.__remove_methods_after_commands()

    def __handle_clearchat(self, message: Message) -> None:
        print(
            f"{message.irc_command} > [{message.channel}] {message.text if message.text is None else ''}"
        )

        self.__execute_methods_after_clearchat(message)
        self.__remove_methods_after_clearchat()

    def __handle_clearmsg(self, message: Message) -> None:
        print(f"{message.irc_command} > [{message.channel}]: {message.text}")

        self.__execute_methods_after_delete_message(message)
        self.__remove_methods_after_delete_message()

    def __handle_globaluserstate(self, message: Message) -> None:
        print(f"{message.irc_command} >")

        self.__execute_methods_after_bot_connected(message)
        self.__remove_methods_after_bot_connected()

    def __handle_hosttarget(self, message: Message) -> None:
        print(f"{message.irc_command} > [{message.channel}]: {message.text}")

        self.__execute_methods_after_toggle_host(message)
        self.__remove_methods_after_toggle_host()

    def __handle_reconnect(self, message: Message) -> None:
        print(f"{message.irc_command} >")

        self.__execute_methods_after_server_reconnect(message)
        self.__remove_methods_after_server_reconnect()

    def __handle_roomstate(self, message: Message) -> None:
        print(f"{message.irc_command} > [{message.channel}]")

        self.__execute_methods_after_channel_change(message)
        self.__remove_methods_after_channel_change()

    def __handle_usernotice(self, message: Message) -> None:
        print(
            f"{message.irc_command} > [{message.channel}]: {message.text if message.text is not None else ''}"
        )

        self.__execute_methods_after_event(message)
        self.__remove_methods_after_event()

    def __handle_userstate(self, message: Message) -> None:
        print(f"{message.irc_command} > [{message.channel}]")

        self.__execute_methods_after_user_join(message)
        self.__remove_methods_after_user_join()

    def __handle_whisper(self, message: Message) -> None:
        print(
            f"{message.irc_command} > {message.irc_args[0] if message.irc_args is not None else ''}: {message.text}"
        )

        self.__execute_methods_after_whisper(message)
        self.__remove_methods_after_whisper()

    def __handle_message(self, received_msg: str) -> None:
        if len(received_msg) == 0:
            return

        message = self.__parse_message(received_msg)

        if message.irc_command == "NOTICE":
            self.__handle_notice(message)

        elif message.irc_command == "PART":
            self.__handle_part(message)

        elif message.irc_command == "PING":
            self.__handle_ping(message)

        elif message.irc_command == "PRIVMSG":
            self.__handle_privmsg(message)

        elif message.irc_command == "CLEARCHAT":
            self.__handle_clearchat(message)

        elif message.irc_command == "CLEARMSG":
            self.__handle_clearmsg(message)

        elif message.irc_command == "GLOBALUSERSTATE":
            self.__handle_globaluserstate(message)

        elif message.irc_command == "HOSTTARGET":
            self.__handle_hosttarget(message)

        elif message.irc_command == "RECONNECT":
            self.__handle_reconnect(message)

        elif message.irc_command == "ROOMSTATE":
            self.__handle_roomstate(message)

        elif message.irc_command == "USERNOTICE":
            self.__handle_usernotice(message)

        elif message.irc_command == "USERSTATE":
            self.__handle_userstate(message)

        elif message.irc_command == "WHISPER":
            self.__handle_whisper(message)

        else:
            print(
                f"{message.irc_command} > [{message.channel}] {message.user}: {message.text}"
            )

    def __loop(self) -> None:
        while not self.__finish:
            try:
                received_msgs = self.irc.recv(2048).decode()

                for received_msg in received_msgs.split("\r\n"):
                    self.__handle_message(received_msg)

            except socket.timeout:
                self.__execute_checks()
                self.__remove_checks()

    def send(self, channel: str, text: str) -> None:
        """
        Sends a message by chat

        Args:
            channel (str): Owner of the chat
            text (str): Message's text
        """

        self.__send_privmsg(channel, text)

    def ban(self, channel: str, user: str, reason: str = "") -> None:
        """
        Bans a user

        Args:
            channel (str): Channel who bans
            username (str): User to ban
            reason (str): Reason of the ban
        """

        self.send(channel, f"/ban @{user} {reason}")

    def unban(self, channel: str, user: str) -> None:
        """
        Undoes the ban of a user

        Args:
            channel (str): Name of the channel who readmits
            user (str): Name of the user to readmit
        """

        self.send(channel, f"/unban @{user}")

    def clear(self, channel: str) -> None:
        """
        Clears the chat

        Args:
            channel (str): Channel to clean the chat
        """

        self.send(channel, "/clear")

    def color(self, channel: str, color: str) -> None:
        """
        Changes the color of the channel's name in the chat

        Args:
            channel (str): Channel to change color
            color (str): New color's name
        """

        self.send(channel, f"/color {color}")

    def commercial(self, channel: str, duration: int = 30) -> None:
        """
        Places advertising in the channel

        Args:
            channel (str): Channel on which start the commercial
            duration (int): Duration of advertising
        """

        self.send(channel, f"/commercial {duration}")

    def delete(self, channel: str, message_id: str) -> None:
        """
        Deletes the specified message from the chat room

        Args:
            channel (str): Channel where the message was sent
            message_id (str): ID of the message
        """

        self.send(channel, f"/delete {message_id}")

    def disconnect(self, channel: str) -> None:
        """
        Closes the session that the command was received from

        Args:
            channel (str): Channel whose session close to
        """

        self.send(channel, "/disconnect")

    def emoteonly(self, channel: str) -> None:
        """
        Activates the "emotes only" mode

        Args:
            channel (str): Channel on which activate the mode
        """

        self.send(channel, "/emoteonly")

    def emoteonlyoff(self, channel: str) -> None:
        """
        Disables "emotes only" mode

        Args:
            channel (str): Channel on which disable the mode
        """

        self.send(channel, "/emoteonlyoff")

    def followers(self, channel: str) -> None:
        """
        Activates the "followers only" mode

        Args:
            channel (str): Channel on which activate the mode
        """

        self.send(channel, "/followers")

    def followersoff(self, channel: str) -> None:
        """
        Disables the "followers only" mode

        Args:
            channel (str): Channel on which disable the mode
        """

        self.send(channel, "/followersoff")

    def help(self, channel: str, command: str = "") -> None:
        """
        Shows detailed information about a command

        Args:
            channel (str): Channel in which show the command's information
            command (str): Command to show information about
        """

        self.send(channel, f"/help {command}")

    def host(self, channel: str, username: str) -> None:
        """
        Hosts a channel

        Args:
            channel (str): Name of the channel who hosts
            username (str): Name of the channel to host
        """

        self.send(channel, f"/host {username}")

    def unhost(self, channel: str) -> None:
        """
        Unhosts the hosted channel

        Args:
            channel (str): Channel who unhosts
        """

        self.send(channel, "/unhost")

    def marker(self, channel: str, description: str = "") -> None:
        """
        Leaves a mark on the channel's stream

        Args:
            channel (str): Channel in which leave the mark
            description (str): Mark's description
        """

        self.send(channel, f"/marker {description}")

    def me(self, channel: str, text: str) -> None:
        """
        Sends a message by chat in italics

        Args:
            channel (str): Owner of the chat
            text (str): Message's text
        """

        self.send(channel, f"/me {text}")

    def mod(self, channel: str, username: str) -> None:
        """
        Makes a user mod

        Args:
            channel (str): Channel who promotes the user
            username (str): Name of the user to be promoted
        """

        self.send(channel, f"/mod {username}")

    def unmod(self, channel: str, username: str) -> None:
        """
        Removes the moderator's rank from a user

        Args:
            channel (str): Channel who removes the moderator's rank
            username (str): User's name
        """

        self.send(channel, f"/unmod {username}")

    def mods(self, channel: str) -> None:
        """
        Shows the moderators list of a channel

        Args:
            channel (str): Channel who owns the moderators
        """

        self.send(channel, "/mods")

    def raid(self, channel: str, username: str) -> None:
        """
        Raids another channel

        Args:
            channel (str): Name of the channel who raids
            username (str): Name of the channel to raid
        """

        self.send(channel, f"/raid {username}")

    def unraid(self, channel: str) -> None:
        """
        Cancels an raid

        Args:
            channel (str): Channel who unraids
        """

        self.send(channel, "/unraid")

    def slow(self, channel: str, duration: int) -> None:
        """
        Activates the "slow" mode

        Args:
            channel (str): Channel on which activate the mode
            duration (int): Time between messages
        """

        self.send(channel, f"/slow {duration}")

    def slowoff(self, channel: str) -> None:
        """
        Disables the "slow" mode

        Args:
            channel (str): Channel on which disable the mode
        """

        self.send(channel, "/slowoff")

    def subscribers(self, channel: str) -> None:
        """
        Activates the "subscribers only" mode

        Args:
            channel (str): Channel on which activate the mode
        """

        self.send(channel, "/subscribers")

    def subscribersoff(self, channel: str) -> None:
        """
        Disables "subscriber only" mode

        Args:
            channel (str): Channel on which disable the mode
        """

        self.send(channel, "/subscribersoff")

    def timeout(self, channel: str, user: str, duration: int = 600) -> None:
        """
        Expels a user temporarily

        Args:
            channel (str): Channel who ejects
            user (str): Name of the user to expel
            duration (int): Ejecting time
        """

        self.send(channel, f"/timeout @{user} {duration}")

    def untimeout(self, channel: str, username: str) -> None:
        """
        Cancels the timeout of a user

        Args:
            channel (str): Channel who ejected the user
            username (str): User to readmit
        """

        self.send(channel, f"/untimeout @{username}")

    def uniquechat(self, channel: str) -> None:
        """
        Activates the "unique" mode

        Args:
            channel (str): Channel on which activate the mode
        """

        self.send(channel, "/uniquechat")

    def uniquechatoff(self, channel: str) -> None:
        """
        Disables the "unique" mode

        Args:
            channel (str): Channel on which disable the mode
        """

        self.send(channel, "/uniquechatoff")

    def vip(self, channel: str, username: str) -> None:
        """
        Makes a user vip

        Args:
            channel (str): Channel who makes a user vip
            username (str): User's name
        """

        self.send(channel, f"/vip {username}")

    def unvip(self, channel: str, username: str) -> None:
        """
        Removes the vip range from a user

        Args:
            channel (str): Channel who remove's the vip range
            username (str): User's name
        """

        self.send(channel, f"/unvip {username}")

    def vips(self, channel: str) -> None:
        """
        Shows the vips list of a channel

        Args:
            channel (str): Channel who owns the vips
        """

        self.send(channel, "/vips")

    def whisper(self, channel: str, user: str, text: str) -> None:
        """
        Whispers to a user

        Args:
            channel (str): Channel on which send the whisp
            user (str): User's name
            text (str): Whisper's text
        """

        self.send(channel, f"/w {user} {text}")

    def add_method_before_join_channel(
        self, name: str, method: Callable[[str], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed before joinnnig a channel

        Args:
            name (str): Method's name
            method (Callable[[str], None]): Method to be executed before joinnnig a channel
        """

        self.custom_methods_before_join_channel[name] = method

    def remove_method_before_join_channel(self, name: str) -> None:
        """
        Removes a method that is executed before joinning a channel

        Args:
            name (str): Method's name
        """

        self.methods_before_join_channel_to_remove.append(name)

    def add_method_after_join_channel(
        self, name: str, method: Callable[[str], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after joinnnig a channel

        Args:
            name (str): Method's name
            method (Callable[[str], None]): Method to be executed after joinnnig a channel
        """

        self.custom_methods_after_join_channel[name] = method

    def remove_method_after_join_channel(self, name: str) -> None:
        """
        Removes a method that is executed after joinning a channel

        Args:
            name (str): Method's name
        """

        self.methods_after_join_channel_to_remove.append(name)

    def add_method_before_leave_channel(
        self, name: str, method: Callable[[str], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed before leaving a channel

        Args:
            name (str): Method's name
            method (Callable[[str], None]): Method to be executed before leaving a channel
        """

        self.custom_methods_before_leave_channel[name] = method

    def remove_method_before_leave_channel(self, name: str) -> None:
        """
        Removes a method that is executed before leaving a channel

        Args:
            name (str): Method's name
        """

        self.methods_before_leave_channel_to_remove.append(name)

    def add_method_after_leave_channel(
        self, name: str, method: Callable[[str], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after leaving a channel

        Args:
            name (str): Method's name
            method (Callable[[str], None]): Method to be executed after leaving a channel
        """

        self.custom_methods_after_leave_channel[name] = method

    def remove_method_after_leave_channel(self, name: str) -> None:
        """
        Removes a method that is executed after leaving a channel

        Args:
            name (str): Method's name
        """

        self.methods_after_leave_channel_to_remove.append(name)

    def add_check(self, name: str, check: Callable[[], None]) -> None:
        """
        Adds a check to the bot
        Checks work permanently

        Args:
            name (str): Check's name
            check (Callable[[], None]): Method that will act as a check
        """

        self.custom_checks[name] = check

    def remove_check(self, name: str) -> None:
        """
        Removes a check from the bot

        Args:
            name (str): Check's name
        """

        self.checks_to_remove.append(name)

    def add_listener(self, name: str, listener: Callable[[Message], None]) -> None:
        """
        Adds a listener to the bot
        Listeners work only when a message is received
        Listeners must receive as a parameter the last message in the chat

        Args:
            name (str): Listener's name
            listener (Callable[[Message], None]): Method that will be executed after every chat message
        """

        self.custom_listeners[name] = listener

    def remove_listener(self, name: str) -> None:
        """
        Removes a listener from the bot

        Args:
            name (str): Listener's name
        """

        self.listeners_to_remove.append(name)

    def add_command(self, name: str, command: Callable[[Message], None]) -> None:
        """
        Adds a command to the bot
        Commands must receive as a parameter the messages which call them

        Args:
            name (str): Command's name
            command (Callable[[Message], None]): Method that will be executed when the command is invoked
        """

        self.custom_commands[name] = command

    def remove_command(self, name: str) -> None:
        """
        Removes a command from the bot

        Args:
            name (str): Command's name
        """

        self.commands_to_remove.append(name)

    def add_method_before_commands(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed before each command

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed before each command
        """

        self.custom_methods_before_commands[name] = method

    def remove_method_before_commands(self, name: str) -> None:
        """
        Removes a method that is executed before each command

        Args:
            name (str): Method's name
        """

        self.methods_before_commands_to_remove.append(name)

    def add_method_after_commands(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after each command

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after each command
        """

        self.custom_methods_after_commands[name] = method

    def remove_method_after_commands(self, name: str) -> None:
        """
        Removes a method that is executed after each command

        Args:
            name (str): Method's name
        """

        self.methods_after_commands_to_remove.append(name)

    def add_method_after_clearchat(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after each chat clearing

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after each chat clearing
        """

        self.custom_methods_after_clearchat[name] = method

    def remove_method_after_clearchat(self, name: str) -> None:
        """
        Removes a method that is executed after each chat clearing

        Args:
            name (str): Method's name
        """

        self.methods_after_clearchat_to_remove.append(name)

    def add_method_after_delete_message(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after each time a message is deleted

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after each time a message is deleted
        """

        self.custom_methods_after_delete_message[name] = method

    def remove_method_after_delete_message(self, name: str) -> None:
        """
        Removes a method that is executed after each time a message is deleted

        Args:
            name (str): Method's name
        """

        self.methods_after_delete_message_to_remove.append(name)

    def add_method_after_bot_connected(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after a bot connects to a chat

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after a bot connects to a chat
        """

        self.custom_methods_after_bot_connected[name] = method

    def remove_method_after_bot_connected(self, name: str) -> None:
        """
        Removes a method that is executed after a bot connects to a chat

        Args:
            name (str): Method's name
        """

        self.methods_after_bot_connected_to_remove.append(name)

    def add_method_after_toggle_host(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after a channel toggles hosting

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after a channel toggles hosting
        """

        self.custom_methods_after_toggle_host[name] = method

    def remove_method_after_toggle_host(self, name: str) -> None:
        """
        Removes a method that is executed after a channel toggles hosting

        Args:
            name (str): Method's name
        """

        self.methods_after_toggle_host_to_remove.append(name)

    def add_method_after_server_reconnect(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after a reconnect warning

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after a reconnect warning
        """

        self.custom_methods_after_server_reconnect[name] = method

    def remove_method_after_server_reconnect(self, name: str) -> None:
        """
        Removes a method that is executed after a reconnect warning

        Args:
            name (str): Method's name
        """

        self.methods_after_server_reconnect_to_remove.append(name)

    def add_method_after_channel_change(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after a channel's chat settings change

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after a channel's chat settings change
        """

        self.custom_methods_after_channel_change[name] = method

    def remove_method_after_channel_change(self, name: str) -> None:
        """
        Removes a method that is executed after a channel's chat settings change

        Args:
            name (str): Method's name
        """

        self.methods_after_channel_change_to_remove.append(name)

    def add_method_after_event(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after an event occurs

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after an event occurs
        """

        self.custom_methods_after_event[name] = method

    def remove_method_after_event(self, name: str) -> None:
        """
        Removes a method that is executed after an event occurs

        Args:
            name (str): Method's name
        """

        self.methods_after_event_to_remove.append(name)

    def add_method_after_user_join(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after an user joins into a channel

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after an user joins into a channel
        """

        self.custom_methods_after_user_join[name] = method

    def remove_method_after_user_join(self, name: str) -> None:
        """
        Removes a method that is executed after an user joins into a channel

        Args:
            name (str): Method's name
        """

        self.methods_after_user_join_to_remove.append(name)

    def add_method_after_whisper(
        self, name: str, method: Callable[[Message], None]
    ) -> None:
        """
        Adds to the bot a method that will be executed after a whisper is received

        Args:
            name (str): Method's name
            method (Callable[[Message], None]): Method to be executed after a whisper is received
        """

        self.custom_methods_after_whisper[name] = method

    def remove_method_after_whisper(self, name: str) -> None:
        """
        Removes a method that is executed after a whisper is received

        Args:
            name (str): Method's name
        """

        self.methods_after_whisper_to_remove.append(name)



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import User


@dataclass
class BannedUser:
    """
    Represents a user that were banned or put in a timeout

    Attributes:
        user (User): The banned user
        expires_at (datetime): The UTC date and time (in RFC3339 format) of when the timeout expires, or an empty string if the user is permanently banned
        created_at (datetime): The UTC date and time (in RFC3339 format) of when the user was banned
        reason (str): The reason the user was banned or put in a timeout if the moderator provided one
        moderator (User): The moderator that banned the user or put them in a timeout
    """

    user: User
    expires_at: datetime
    created_at: datetime
    reason: str
    moderator: User



from dataclasses import dataclass
from datetime import datetime


@dataclass
class Transport:
    """
    Represents a transport for sending notifications

    Attributes:
        method (str): The transport method
            Possible values are: webhook, websocket, conduit
        callback (str): The callback URL where the notifications are sent
        session_id (str): An ID that identifies the WebSocket that notifications are sent to
        connected_at (datetime): The UTC date and time that the WebSocket connection was established
        disconnected_at (datetime | None): The UTC date and time that the WebSocket connection was lost
        conduit_id (str | None): An ID that identifies the conduit to send notifications to
    """

    method: str
    callback: str
    session_id: str
    connected_at: datetime
    disconnected_at: datetime | None = None
    conduit_id: str | None = None



from datetime import datetime

from .._utils import date, http
from ..dataclasses import Channel, Prediction, PredictionOutcome, Predictor, User

ENDPOINT_PREDICTIONS = "https://api.twitch.tv/helix/predictions"


def get_predictions(
    token: str,
    client_id: str,
    broadcaster_id: str,
    prediction_ids: list[str] | None = None,
    first: int = 20,
) -> list[Prediction]:
    url = ENDPOINT_PREDICTIONS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}
    params["broadcaster_id"] = broadcaster_id

    if prediction_ids is not None and len(prediction_ids) > 0:
        params["id"] = prediction_ids

    predictions = http.send_get_with_pagination(url, headers, params, first, 20)

    return [
        Prediction(
            prediction["id"],
            Channel(
                User(
                    prediction["broadcaster_id"],
                    prediction["broadcaster_login"],
                    prediction["broadcaster_name"],
                )
            ),
            prediction["title"],
            prediction["winning_outcome_id"],
            [
                PredictionOutcome(
                    outcome["id"],
                    outcome["title"],
                    outcome["users"],
                    outcome["channel_points"],
                    [
                        Predictor(
                            User(
                                predictor["user_id"],
                                predictor["user_login"],
                                predictor["user_name"],
                            ),
                            predictor["channel_points_used"],
                            predictor["channel_points_won"],
                        )
                        for predictor in outcome["top_predictors"]
                    ],
                    outcome["color"],
                )
                for outcome in prediction["outcomes"]
            ],
            prediction["prediction_window"],
            prediction["status"],
            datetime.strptime(prediction["created_at"], date.RFC3339_FORMAT),
            datetime.strptime(prediction["ended_at"], date.RFC3339_FORMAT),
            datetime.strptime(prediction["locked_at"], date.RFC3339_FORMAT),
        )
        for prediction in predictions
    ]


def create_prediction(
    token: str,
    client_id: str,
    broadcaster_id: str,
    title: str,
    outcomes: list[str],
    prediction_window: int,
) -> Prediction:
    url = ENDPOINT_PREDICTIONS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
        "Content-Type": "application/json",
    }
    payload = {
        "broadcaster_id": broadcaster_id,
        "title": title,
        "prediction_window": prediction_window,
    }

    outcomes_payload = []

    for outcome in outcomes:
        outcomes_payload.append({"title": outcome})

    payload["outcomes"] = outcomes_payload

    prediction = http.send_post_get_result(url, headers, payload)[0]

    return Prediction(
        prediction["id"],
        Channel(
            User(
                prediction["broadcaster_id"],
                prediction["broadcaster_login"],
                prediction["broadcaster_name"],
            )
        ),
        prediction["title"],
        prediction["winning_outcome_id"],
        [
            PredictionOutcome(
                outcome["id"],
                outcome["title"],
                outcome["users"],
                outcome["channel_points"],
                [
                    Predictor(
                        User(
                            predictor["user_id"],
                            predictor["user_login"],
                            predictor["user_name"],
                        ),
                        predictor["channel_points_used"],
                        predictor["channel_points_won"],
                    )
                    for predictor in outcome["top_predictors"]
                ],
                outcome["color"],
            )
            for outcome in prediction["outcomes"]
        ],
        prediction["prediction_window"],
        prediction["status"],
        datetime.strptime(prediction["created_at"], date.RFC3339_FORMAT),
        datetime.strptime(prediction["ended_at"], date.RFC3339_FORMAT),
        datetime.strptime(prediction["locked_at"], date.RFC3339_FORMAT),
    )


def end_prediction(
    token: str,
    client_id: str,
    broadcaster_id: str,
    prediction_id: str,
    status: str,
    winning_outcome_id: str = "",
) -> Prediction:
    url = ENDPOINT_PREDICTIONS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"broadcaster_id": broadcaster_id, "id": prediction_id, "status": status}

    if winning_outcome_id != "":
        data["winning_outcome_id"] = winning_outcome_id

    prediction = http.send_patch_get_result(url, headers, data)[0]

    return Prediction(
        prediction["id"],
        Channel(
            User(
                prediction["broadcaster_id"],
                prediction["broadcaster_login"],
                prediction["broadcaster_name"],
            )
        ),
        prediction["title"],
        prediction["winning_outcome_id"],
        [
            PredictionOutcome(
                outcome["id"],
                outcome["title"],
                outcome["users"],
                outcome["channel_points"],
                [
                    Predictor(
                        User(
                            predictor["user_id"],
                            predictor["user_login"],
                            predictor["user_name"],
                        ),
                        predictor["channel_points_used"],
                        predictor["channel_points_won"],
                    )
                    for predictor in outcome["top_predictors"]
                ],
                outcome["color"],
            )
            for outcome in prediction["outcomes"]
        ],
        prediction["prediction_window"],
        prediction["status"],
        datetime.strptime(prediction["created_at"], date.RFC3339_FORMAT),
        datetime.strptime(prediction["ended_at"], date.RFC3339_FORMAT),
        datetime.strptime(prediction["locked_at"], date.RFC3339_FORMAT),
    )



from dataclasses import dataclass


@dataclass
class ContentClassificationLabel:
    """
    Represents a content classification label

    Attributes:
        label_id (str): Unique identifier for the label
        description (str): Localized description of the label
        name (str): Localized name of the label
    """

    label_id: str
    description: str
    name: str



from datetime import datetime

from .._utils import date, http
from ..dataclasses import Channel, User, Video


def get_videos(
    token: str,
    client_id: str,
    video_ids: list[str] | None = None,
    user_id: str | None = None,
    game_id: str | None = None,
    first: int = 20,
    language: str | None = None,
    period: str = "all",
    sort: str = "time",
    video_type: str = "all",
) -> list[Video]:
    url = "https://api.twitch.tv/helix/videos"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}
    params["period"] = period
    params["sort"] = sort
    params["type"] = video_type

    if video_ids is not None and len(video_ids) > 0:
        params["id"] = video_ids

    if user_id is not None:
        params["user_id"] = user_id

    if game_id is not None:
        params["game_id"] = game_id

    if language is not None:
        params["language"] = language

    videos = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        Video(
            video["id"],
            video["stream_id"],
            Channel(User(video["user_id"], video["user_login"], video["user_name"])),
            video["title"],
            video["description"],
            datetime.strptime(video["created_at"], date.RFC3339_FORMAT),
            datetime.strptime(video["published_at"], date.RFC3339_FORMAT),
            video["url"],
            video["thumbnail_url"],
            video["viewable"],
            video["view_count"],
            video["language"],
            video["type"],
            video["duration"],
            [
                (segment["duration"], segment["offset"])
                for segment in video["muted_segments"]
            ],
        )
        for video in videos
    ]


def delete_video(token: str, client_id: str, video_id: str) -> None:
    url = "https://api.twitch.tv/helix/videos"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"id": video_id}

    http.send_delete(url, headers, data)



from dataclasses import dataclass
from datetime import datetime


@dataclass
class StreamMarker:
    """
    Represents a marker in a stream

    Attributes:
        marker_id (str): An ID that identifies this marker
        created_at (datetime): The UTC date and time (in RFC3339 format) of when the user created the marker
        position_seconds (int): The relative offset (in seconds) of the marker from the beginning of the stream
        description (str): A description that the user gave the marker to help them remember why they marked the location
    """

    marker_id: str
    created_at: datetime
    position_seconds: int
    description: str



from dataclasses import dataclass


@dataclass
class Conduit:
    """
    Represents a conduit

    Attributes:
        conduit_id (str): Conduit ID
        shard_count (int): Number of shards associated with this conduit
    """

    conduit_id: str
    shard_count: int



from dataclasses import dataclass
from datetime import datetime


@dataclass
class GameAnalyticsReport:
    """
    Represents an analytics report for a game

    Attributes:
        game_id (str): An ID that identifies the game that the report was generated for
        url (str): The URL that you use to download the report
        type (str): The type of report
        started_at (datetime): The reporting window’s start date
        ended_at (datetime): The reporting window’s end date
    """

    game_id: str
    url: str
    type: str
    started_at: datetime
    ended_at: datetime



from datetime import datetime

from .._utils import date, http
from ..dataclasses import Channel, CreatorGoal, User


def get_creator_goals(
    token: str, client_id: str, broadcaster_id: str
) -> list[CreatorGoal]:
    url = "https://api.twitch.tv/helix/goals"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id}

    goals = http.send_get(url, headers, params)

    return [
        CreatorGoal(
            goal["id"],
            Channel(
                User(
                    goal["broadcaster_id"],
                    goal["broadcaster_login"],
                    goal["broadcaster_name"],
                )
            ),
            goal["type"],
            goal["description"],
            goal["current_amount"],
            goal["target_amount"],
            datetime.strptime(goal["created_at"], date.RFC3339_FORMAT),
        )
        for goal in goals
    ]



from dataclasses import dataclass


@dataclass
class HypeTrainContribution:
    """
    Represents a contribution to a Hype Train's goal

    Attributes:
        total (int): The total amount contributed
        type (str): The contribution method used
            Possible values: BITS SUBS OTHER
        user (str): The ID of the user that made the contribution
    """

    total: int
    type: str
    user: str



from dataclasses import dataclass


@dataclass
class Message:
    """
    Represents a message

    Attributes:
        prefix (str | None): Message's prefix
        user (str | None): User who has sent the message
        channel (str | None): Channel on which the message was sent
        irc_command (str | None): IRC command related to the message
        irc_args (list[str] | None): IRC command's arguments
        text (str | None): Message's text
        text_command (str | None): Command related to the message
        text_args (list[str] | None): Command's arguments
    """

    prefix: str | None = None
    user: str | None = None
    channel: str | None = None
    irc_command: str | None = None
    irc_args: list[str] | None = None
    text: str | None = None
    text_command: str | None = None
    text_args: list[str] | None = None



from dataclasses import dataclass
from datetime import datetime


@dataclass
class ExtensionSecret:
    """
    Represents an extension's secret

    Attributes:
        content (str): The raw secret that you use with JWT encoding
        active_at (datetime): The UTC date and time (in RFC3339 format) that you may begin using this secret to sign a JWT
        expires_at (datetime): The UTC date and time (in RFC3339 format) that you must stop using this secret to decode a JWT
    """

    content: str
    active_at: datetime
    expires_at: datetime



from datetime import datetime

from .._utils import date, http
from ..dataclasses import (
    Channel,
    Extension,
    ExtensionConfigurationSegment,
    ExtensionSecret,
    ExtensionTransaction,
    Game,
    Product,
    ProductCost,
    User,
)


def get_extension_transactions(
    token: str,
    client_id: str,
    extension_id: str,
    transaction_ids: list[str] | None = None,
    first: int = 20,
) -> list[ExtensionTransaction]:
    url = "https://api.twitch.tv/helix/extensions/transactions"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}
    params["extension_id"] = extension_id

    if transaction_ids is not None and len(transaction_ids) > 0:
        params["id"] = transaction_ids

    transactions = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        ExtensionTransaction(
            transaction["id"],
            datetime.strptime(transaction["timestamp"], date.RFC3339_FORMAT),
            Channel(
                User(
                    transaction["broadcaster_id"],
                    transaction["broadcaster_login"],
                    transaction["broadcaster_name"],
                )
            ),
            User(
                transaction["user_id"],
                transaction["user_login"],
                transaction["user_name"],
            ),
            transaction["product_type"],
            Product(
                transaction["product_data"]["sku"],
                ProductCost(
                    transaction["product_data"]["cost"]["amount"],
                    transaction["product_data"]["cost"]["type"],
                ),
                transaction["product_data"]["inDevelopment"],
                transaction["product_data"]["displayName"],
                transaction["product_data"]["expiration"],
                transaction["product_data"]["broadcast"],
                transaction["product_data"]["domain"],
            ),
        )
        for transaction in transactions
    ]


def get_extension_configuration_segment(
    token: str, client_id: str, broadcaster_id: str, extension_id: str, segment: str
) -> ExtensionConfigurationSegment:
    url = "https://api.twitch.tv/helix/extensions/configurations"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {
        "broadcaster_id": broadcaster_id,
        "extension_id": extension_id,
        "segment": segment,
    }

    configuration_segment = http.send_get(url, headers, params)[0]

    return ExtensionConfigurationSegment(
        configuration_segment["segment"],
        configuration_segment["broadcaster_id"],
        configuration_segment["content"],
        configuration_segment["version"],
    )


def set_extension_configuration_segment(
    token: str,
    client_id: str,
    extension_id: str,
    segment: str,
    broadcaster_id: str | None = None,
    content: str | None = None,
    version: str | None = None,
) -> None:
    url = "https://api.twitch.tv/helix/extensions/configurations"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"extension_id": extension_id, "segment": segment}

    if broadcaster_id is not None:
        data["broadcaster_id"] = broadcaster_id

    if content is not None:
        data["content"] = content

    if version is not None:
        data["version"] = version

    http.send_put(url, headers, data)


def set_extension_required_configuration(
    token: str,
    client_id: str,
    broadcaster_id: str,
    extension_id: str,
    extension_version: str,
    configuration_version: str,
) -> None:
    url = "https://api.twitch.tv/helix/extensions/required_configuration"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {
        "broadcaster_id": broadcaster_id,
        "extension_id": extension_id,
        "extension_version": extension_version,
        "configuration_version": configuration_version,
    }

    http.send_put(url, headers, data)


def send_extension_pubsub_message(
    token: str,
    client_id: str,
    target: list[str],
    broadcaster_id: str,
    is_global_broadcast: bool,
    message: str,
) -> None:
    url = "https://api.twitch.tv/helix/extensions/pubsub"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    payload = {
        "target": target,
        "broadcaster_id": broadcaster_id,
        "is_global_broadcast": is_global_broadcast,
        "message": message,
    }

    http.send_post(url, headers, payload)


def get_extension_live_channels(
    token: str, client_id: str, extension_id: str, first: int = 20
) -> list[Channel]:
    url = "https://api.twitch.tv/helix/extensions/live"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": {client_id},
    }
    params = {"extension_id": extension_id}

    channels = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        Channel(
            User(
                channel["broadcaster_id"],
                channel["broadcaster_name"].lower(),
                channel["broadcaster_name"],
            ),
            game=Game(channel["game_id"], channel["game_name"]),
            title=channel["title"],
        )
        for channel in channels
    ]


def get_extension_secrets(
    token: str, client_id: str
) -> list[tuple[str, list[ExtensionSecret]]]:
    url = "https://api.twitch.tv/helix/extensions/jwt/secrets"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }

    extension_secrets = http.send_get(url, headers, {})

    return [
        (
            extension_secret["format_version"],
            [
                ExtensionSecret(
                    secret["content"], secret["active_at"], secret["expires_at"]
                )
                for secret in extension_secret["secrets"]
            ],
        )
        for extension_secret in extension_secrets
    ]


def create_extension_secret(
    token: str, client_id: str, delay: int = 300
) -> list[tuple[str, list[ExtensionSecret]]]:
    url = "https://api.twitch.tv/helix/extensions/jwt/secrets"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    payload = {"delay": delay}

    extension_secrets = http.send_post_get_result(url, headers, payload)

    return [
        (
            extension_secret["format_version"],
            [
                ExtensionSecret(
                    secret["content"], secret["active_at"], secret["expires_at"]
                )
                for secret in extension_secret["secrets"]
            ],
        )
        for extension_secret in extension_secrets
    ]


def send_extension_chat_message(
    token: str,
    client_id: str,
    broadcaster_id: str,
    text: str,
    extension_id: str,
    extension_version: str,
) -> None:
    url = "https://api.twitch.tv/helix/extensions/chat"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    payload = {
        "broadcaster_id": broadcaster_id,
        "text": text,
        "extension_id": extension_id,
        "extension_version": extension_version,
    }

    http.send_post(url, headers, payload)


def get_extensions(
    token: str, client_id: str, extension_id: str, extension_version: str | None = None
) -> Extension:
    url = "https://api.twitch.tv/helix/extensions"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"extension_id": extension_id}

    if extension_version is not None:
        params["extension_version"] = extension_version

    extension = http.send_get(url, headers, params)[0]

    return Extension(
        extension["author_name"],
        extension["bits_enables"],
        extension["can_install"],
        extension["configuration_location"],
        extension["description"],
        extension["eula_tos_url"],
        extension["has_chat_support"],
        extension["icon_url"],
        extension["icon_urls"],
        extension["id"],
        extension["name"],
        extension["privacy_policy_url"],
        extension["request_identity_link"],
        extension["screenshot_urls"],
        extension["state"],
        extension["subscriptions_support_level"],
        extension["summary"],
        extension["support_email"],
        extension["version"],
        extension["viewer_summary"],
        extension["views"],
        extension["allowlisted_config_urls"],
        extension["allowlisted_panel_urls"],
    )


def get_released_extensions(
    token: str, client_id: str, extension_id: str, extension_version: str | None = None
) -> Extension:
    url = "https://api.twitch.tv/helix/extensions/released"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"extension_id": extension_id}

    if extension_version is not None:
        params["extension_version"] = extension_version

    extension = http.send_get(url, headers, params)[0]

    return Extension(
        extension["author_name"],
        extension["bits_enables"],
        extension["can_install"],
        extension["configuration_location"],
        extension["description"],
        extension["eula_tos_url"],
        extension["has_chat_support"],
        extension["icon_url"],
        extension["icon_urls"],
        extension["id"],
        extension["name"],
        extension["privacy_policy_url"],
        extension["request_identity_link"],
        extension["screenshot_urls"],
        extension["state"],
        extension["subscriptions_support_level"],
        extension["summary"],
        extension["support_email"],
        extension["version"],
        extension["viewer_summary"],
        extension["views"],
        extension["allowlisted_config_urls"],
        extension["allowlisted_panel_urls"],
    )


def get_extension_bits_products(
    token: str, extension_client_id: str, should_include_all: bool = False
) -> list[Product]:
    url = "https://api.twitch.tv/helix/bits/extensions"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": extension_client_id,
    }
    params = {"should_include_all": should_include_all}

    products = http.send_get(url, headers, params)

    return [
        Product(
            product["sku"],
            ProductCost(product["cost"]["amount"], product["cost"]["type"]),
            product["in_development"],
            product["display_name"],
            product["expiration"],
            product["is_broadcast"],
        )
        for product in products
    ]


def update_extension_bits_product(
    token: str,
    extension_client_id: str,
    sku: str,
    cost: dict,
    display_name: str,
    in_development: bool | None = None,
    expiration: str | None = None,
    is_broadcast: bool | None = None,
) -> list[Product]:
    url = "https://api.twitch.tv/helix/bits/extensions"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": extension_client_id,
    }
    data = {"sku": sku, "cost": cost, "display_name": display_name}

    if in_development is not None:
        data["in_development"] = in_development

    if expiration is not None:
        data["expiration"] = expiration

    if is_broadcast is not None:
        data["is_broadcast"] = is_broadcast

    products = http.send_put_get_result(url, headers, data)

    return [
        Product(
            product["sku"],
            ProductCost(product["cost"]["amount"], product["cost"]["type"]),
            product["in_development"],
            product["display_name"],
            product["expiration"],
            product["is_broadcast"],
        )
        for product in products
    ]



from .._utils import http
from ..dataclasses import Badge, BadgeVersion, ChatSettings, Emote, User


def get_chatters(
    token: str, client_id: str, broadcaster_id: str, moderator_id: str, first: int = 100
) -> list[User]:
    url = "https://api.twitch.tv/helix/chat/chatters"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id, "moderator_id": moderator_id}

    chatters = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        User(chatter["user_id"], chatter["user_login"], chatter["user_name"])
        for chatter in chatters
    ]


def get_channel_emotes(token: str, client_id: str, broadcaster_id: str) -> list[Emote]:
    url = "https://api.twitch.tv/helix/chat/emotes"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id}

    emotes = http.send_get(url, headers, params)

    return [
        Emote(
            emote["id"],
            emote["name"],
            emote["format"],
            emote["scale"],
            emote["theme_mode"],
            emote["images"],
            emote["tier"],
            emote["emote_type"],
            emote["emote_set_id"],
        )
        for emote in emotes
    ]


def get_global_emotes(token: str, client_id: str) -> list[Emote]:
    url = "https://api.twitch.tv/helix/chat/emotes/global"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }

    emotes = http.send_get(url, headers, {})

    return [
        Emote(
            emote["id"],
            emote["name"],
            emote["format"],
            emote["scale"],
            emote["theme_mode"],
            emote["images"],
        )
        for emote in emotes
    ]


def get_emote_sets(token: str, client_id: str, emote_set_id: list[str]) -> list[Emote]:
    url = "https://api.twitch.tv/helix/chat/emotes/set"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"emote_set_id": emote_set_id}

    emotes = http.send_get(url, headers, params)

    return [
        Emote(
            emote["id"],
            emote["name"],
            emote["format"],
            emote["scale"],
            emote["theme_mode"],
            emote["images"],
            emote_type=emote["emote_type"],
            emote_set_id=emote["emote_set_id"],
            owner_id=emote["owner_id"],
        )
        for emote in emotes
    ]


def get_channel_chat_badges(
    token: str, client_id: str, broadcaster_id: str
) -> list[Badge]:
    url = "https://api.twitch.tv/helix/chat/badges"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id}

    badges = http.send_get(url, headers, params)

    return [
        Badge(
            badge["set_id"],
            [
                BadgeVersion(
                    version["id"],
                    version["image_url_1x"],
                    version["image_url_2x"],
                    version["image_url_4x"],
                    version["title"],
                    version["description"],
                    version["click_action"],
                    version["click_url"],
                )
                for version in badge["versions"]
            ],
        )
        for badge in badges
    ]


def get_global_chat_badges(token: str, client_id: str) -> list[Badge]:
    url = "https://api.twitch.tv/helix/chat/badges/global"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }

    badges = http.send_get(url, headers, {})

    return [
        Badge(
            badge["set_id"],
            [
                BadgeVersion(
                    version["id"],
                    version["image_url_1x"],
                    version["image_url_2x"],
                    version["image_url_4x"],
                    version["title"],
                    version["description"],
                    version["click_action"],
                    version["click_url"],
                )
                for version in badge["versions"]
            ],
        )
        for badge in badges
    ]


def get_chat_settings(
    token: str, client_id: str, broadcaster_id: str, moderator_id: str | None = None
) -> ChatSettings:
    url = "https://api.twitch.tv/helix/chat/settings"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id}

    if moderator_id is not None:
        params["moderator_id"] = moderator_id

    settings = http.send_get(url, headers, params)[0]

    return ChatSettings(
        settings["broadcaster_id"],
        settings["emote_mode"],
        settings["follower_mode"],
        settings["follower_mode_duration"],
        settings["moderator_id"],
        settings["non_moderator_chat_delay"],
        settings["non_moderator_chat_delay_duration"],
        settings["slow_mode"],
        settings["slow_mode_wait_time"],
        settings["subscriber_mode"],
        settings["unique_chat_mode"],
    )


def get_user_emotes(
    token: str, client_id: str, user_id: str, broadcaster_id: str | None = None
) -> list[Emote]:
    url = "https://api.twitch.tv/helix/chat/emotes/user"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"user_id": user_id}

    if broadcaster_id is not None:
        params["broadcaster_id"] = broadcaster_id

    emotes = http.send_get_with_infinite_pagination(url, headers, params)

    return [
        Emote(
            emote["id"],
            emote["name"],
            emote["format"],
            emote["scale"],
            emote["theme_mode"],
            emote_type=emote["emote_type"],
            emote_set_id=emote["emote_set_id"],
            owner_id=emote["owner_id"],
        )
        for emote in emotes
    ]


def update_chat_settings(
    token: str,
    client_id: str,
    broadcaster_id: str,
    moderator_id: str,
    emote_mode: bool | None = None,
    follower_mode: bool | None = None,
    follower_mode_duration: int | None = None,
    non_moderator_chat_delay: bool | None = None,
    non_moderator_chat_delay_duration: int | None = None,
    slow_mode: bool | None = None,
    slow_mode_wait_time: int | None = None,
    subscriber_mode: bool | None = None,
    unique_chat_mode: bool | None = None,
) -> ChatSettings:
    url = "https://api.twitch.tv/helix/chat/settings"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {}
    data["broadcaster_id"] = broadcaster_id
    data["moderator_id"] = moderator_id

    if emote_mode is not None:
        data["emote_mode"] = emote_mode

    if follower_mode is not None:
        data["follower_mode"] = follower_mode

    if follower_mode_duration is not None:
        data["follower_mode_duration"] = follower_mode_duration

    if non_moderator_chat_delay is not None:
        data["non_moderator_chat_delay"] = non_moderator_chat_delay

    if non_moderator_chat_delay_duration is not None:
        data["non_moderator_chat_delay_duration"] = non_moderator_chat_delay_duration

    if slow_mode is not None:
        data["slow_mode"] = slow_mode

    if slow_mode_wait_time is not None:
        data["slow_mode_wait_time"] = slow_mode_wait_time

    if subscriber_mode is not None:
        data["subscriber_mode"] = subscriber_mode

    if unique_chat_mode is not None:
        data["unique_chat_mode"] = unique_chat_mode

    settings = http.send_patch_get_result(url, headers, data)[0]

    return ChatSettings(
        settings["broadcaster_id"],
        settings["emote_mode"],
        settings["follower_mode"],
        settings["follower_mode_duration"],
        settings["moderator_id"],
        settings["non_moderator_chat_delay"],
        settings["non_moderator_chat_delay_duration"],
        settings["slow_mode"],
        settings["slow_mode_wait_time"],
        settings["subscriber_mode"],
        settings["unique_chat_mode"],
    )


def send_chat_announcement(
    token: str,
    client_id: str,
    broadcaster_id: str,
    moderator_id: str,
    message: str,
    color: str | None = None,
) -> None:

    url = "https://api.twitch.tv/helix/chat/announcements"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
        "Content-Type": "application/json",
    }
    payload = {
        "broadcaster_id": broadcaster_id,
        "moderator_id": moderator_id,
        "message": message,
    }

    if color is not None:
        payload["color"] = color

    http.send_post(url, headers, payload)


def send_a_shoutout(
    token: str,
    client_id: str,
    from_broadcaster_id: str,
    to_broadcaster_id: str,
    moderator_id: str,
) -> None:
    url = "https://api.twitch.tv/helix/chat/shoutouts"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    payload = {
        "from_broadcaster_id": from_broadcaster_id,
        "to_broadcaster_id": to_broadcaster_id,
        "moderator_id": moderator_id,
    }

    http.send_post(url, headers, payload)


def send_chat_message(
    token: str,
    client_id: str,
    broadcaster_id: str,
    sender_id: str,
    message: str,
    reply_parent_message_id: str | None = None,
) -> dict:
    url = "https://api.twitch.tv/helix/chat/messages"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
        "Content-Type": "application/json",
    }
    payload = {
        "broadcaster_id": broadcaster_id,
        "sender_id": sender_id,
        "message": message,
    }

    if reply_parent_message_id is not None:
        payload["reply_parent_message_id"] = reply_parent_message_id

    return http.send_post_get_result(url, headers, payload)[0]


def get_user_chat_color(
    token: str, client_id: str, user_id: list[str]
) -> list[tuple[User, str]]:
    url = "https://api.twitch.tv/helix/chat/color"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"user_id": user_id}

    color_settings = http.send_get(url, headers, params)

    return [
        (
            User(setting["user_id"], setting["user_login"], setting["user_name"]),
            setting["color"],
        )
        for setting in color_settings
    ]


def update_user_chat_color(
    token: str, client_id: str, user_id: str, color: str
) -> None:
    url = "https://api.twitch.tv/helix/chat/color"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"user_id": user_id, "color": color}

    http.send_put(url, headers, data)



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Channel


@dataclass
class Video:
    """
    Represents a video

    Attributes:
        video_id (str): ID of the video
        stream_id (str): The ID of the stream that the video originated from
        channel (Channel): The channel that owns the video
        title (str): Title of the video
        description (str): Description of the video
        created_at (datetime): Date of creation of the video
        published_at (datetime): Date of publication of the video
        url (str): URL of the video
        thumbnail_url (str): URL of the preview image of the video
        viewable (str): Indicates whether the video is publicly viewable
        view_count (int): Number of times the video has been viewed
        language (str): Language of the video
        type (str): Type of the video
        duration (str): Duration of the video
        muted_segments (list[tuple[int, int]]): The segments that Twitch Audio Recognition muted
    """

    video_id: str
    stream_id: str
    channel: Channel
    title: str
    description: str
    created_at: datetime
    published_at: datetime
    url: str
    thumbnail_url: str
    viewable: str
    view_count: int
    language: str
    video_type: str
    duration: str
    muted_segments: list[tuple[int, int]]



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import HypeTrainEventData


@dataclass
class HypeTrainEvent:
    """
    Represents a Hype Train event

    Attributes:
        event_id (str): The distinct ID of the event
        event_type (str): Displays hypetrain.{event_name}, currently only hypetrain.progression
        event_timestamp (datetime): RFC3339 formatted timestamp of event
        version (str): Returns the version of the endpoint
        event_data (dict): The event data
    """

    event_id: str
    event_type: str
    event_timestamp: datetime
    version: str
    event_data: HypeTrainEventData



from dataclasses import dataclass

from ..dataclasses import CharityCampaignAmount, User


@dataclass
class CharityCampaignDonation:
    """
    Represents a donation made to a charity campaign

    Attributes:
        donation_id (str): An ID that identifies the donation
        campaign_id (str): An ID that identifies the charity campaign that the donation applies to
        user (User): The user that donated money to the campaign
        amount (CharityCampaignAmount): The amount of money that the user donated
    """

    donation_id: str
    campaign_id: str
    user: User
    amount: CharityCampaignAmount



from .._utils import http


def send_whisper(
    token: str, client_id: str, from_user_id: str, to_user_id: str, message: str
) -> None:
    url = "https://api.twitch.tv/helix/whispers"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
        "Content-Type": "application/json",
    }
    payload = {
        "from_user_id": from_user_id,
        "to_user_id": to_user_id,
        "message": message,
    }

    http.send_post(url, headers, payload)



from dataclasses import dataclass

from ..dataclasses import Guest


@dataclass
class GuestStarSession:
    """
    Represents a Guest Star Session

    Attributes:
        guest_star_session_id (str): ID uniquely representing the Guest Star session
        guests (list[Guest]): List of guests currently interacting with the Guest Star session
    """

    guest_star_session_id: str
    guests: list[Guest]



from .ad_schedule import AdSchedule
from .automod_settings import AutoModSettings
from .badge_version import BadgeVersion
from .badge import Badge
from .user import User
from .banned_user import BannedUser
from .bits_leaderboard_leader import BitsLeaderboardLeader
from .blocked_term import BlockedTerm
from .game import Game
from .channel import Channel
from .charity_campaign_amount import CharityCampaignAmount
from .charity_campaign import CharityCampaign
from .charity_campaign_donation import CharityCampaignDonation
from .chat_settings import ChatSettings
from .cheermote_tier import CheermoteTier
from .cheermote import Cheermote
from .clip import Clip
from .commercial import Commercial
from .conduit import Conduit
from .transport import Transport
from .conduit_shard import ConduitShard
from .content_classification_label import ContentClassificationLabel
from .creator_goal import CreatorGoal
from .drop_entitlement import DropEntitlement
from .emote import Emote
from .eventsub_subscription import EventSubSubscription
from .extension import Extension
from .extension_analytics_report import ExtensionAnalyticsReport
from .extension_configuration_segment import ExtensionConfigurationSegment
from .extension_secret import ExtensionSecret
from .product_cost import ProductCost
from .product import Product
from .extension_transaction import ExtensionTransaction
from .game_analytics_report import GameAnalyticsReport
from .guest import Guest
from .guest_star_invite import GuestStarInvite
from .guest_star_session import GuestStarSession
from .guest_star_settings import GuestStarSettings
from .hype_train_contribution import HypeTrainContribution
from .hype_train_event_data import HypeTrainEventData
from .hypetrain_event import HypeTrainEvent
from .message import Message
from .poll_choice import PollChoice
from .poll import Poll
from .predictor import Predictor
from .prediction_outcome import PredictionOutcome
from .prediction import Prediction
from .reward import Reward
from .redemption import Redemption
from .shield_mode_status import ShieldModeStatus
from .stream import Stream
from .stream_marker import StreamMarker
from .stream_schedule_segment import StreamScheduleSegment
from .stream_schedule import StreamSchedule
from .subscription import Subscription
from .tag import Tag
from .team import Team
from .unban_request import UnbanRequest
from .video import Video



from dataclasses import dataclass
from datetime import datetime


@dataclass
class BlockedTerm:
    """
    Represents a blocked term

    Attributes:
        broadcaster_id (str): The broadcaster that owns the list of blocked terms
        moderator_id (str): The moderator that blocked the word or phrase from being used in the broadcaster’s chat room
        term_id (str): An ID that identifies this blocked term
        text (str): The blocked word or phrase
        created_at (datetime): The UTC date and time (in RFC3339 format) that the term was blocked
        updated_at (datetime): The UTC date and time (in RFC3339 format) that the term was updated
        expires_at (datetime): The UTC date and time (in RFC3339 format) that the blocked term is set to expire
    """

    broadcaster_id: str
    moderator_id: str
    term_id: str
    text: str
    created_at: datetime
    updated_at: datetime
    expires_at: datetime



from datetime import datetime

from .._utils import date, http
from ..dataclasses import Channel, Game, StreamSchedule, StreamScheduleSegment, User

ENDPOINT_SEGMENTS = "https://api.twitch.tv/helix/schedule/segment"


def get_channel_stream_schedule(
    token: str,
    client_id: str,
    broadcaster_id: str,
    stream_segment_id: list[str] | None = None,
    start_time: datetime | None = None,
    first: int = 20,
) -> list[StreamSchedule]:
    url = "https://api.twitch.tv/helix/schedule"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}
    params["broadcaster_id"] = broadcaster_id

    if stream_segment_id is not None and len(stream_segment_id) > 0:
        params["id"] = stream_segment_id

    if start_time is not None:
        params["start_time"] = start_time

    schedules = http.send_get_with_pagination(url, headers, params, first, 25)

    return [
        StreamSchedule(
            [
                StreamScheduleSegment(
                    segment["id"],
                    datetime.strptime(segment["start_time"], date.RFC3339_FORMAT),
                    datetime.strptime(segment["end_time"], date.RFC3339_FORMAT),
                    segment["title"],
                    datetime.strptime(segment["canceled_until"], date.RFC3339_FORMAT),
                    Game(segment["category"]["id"], segment["category"]["name"]),
                    segment["is_recurring"],
                )
                for segment in schedule["segments"]
            ],
            Channel(
                User(
                    schedule["broadcaster_id"],
                    schedule["broadcaster_login"],
                    schedule["broadcaster_name"],
                )
            ),
            (
                datetime.strptime(
                    schedule["vacation"]["start_time"], date.RFC3339_FORMAT
                ),
                datetime.strptime(
                    schedule["vacation"]["end_time"], date.RFC3339_FORMAT
                ),
            ),
        )
        for schedule in schedules
    ]


def get_channel_icalendar(broadcaster_id: str) -> str:
    """
    Gets all scheduled broadcasts from a channel’s stream schedule as an iCalendar

    Args:
        broadcaster_id (str): User ID of the broadcaster who owns the channel streaming schedule

    Returns:
        str
    """

    url = "https://api.twitch.tv/helix/schedule/icalendar"
    params = {"broadcaster_id": broadcaster_id}

    return http.send_get_text(url, params)


def update_channel_stream_schedule(
    token: str,
    client_id: str,
    broadcaster_id: str,
    is_vacation_enabled: bool | None = None,
    vacation_start_time: datetime | None = None,
    vacation_end_time: datetime | None = None,
    timezone: str | None = None,
) -> None:
    url = "https://api.twitch.tv/helix/schedule/settings"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {}
    data["broadcaster_id"] = broadcaster_id

    if is_vacation_enabled is not None:
        data["is_vacation_enabled"] = is_vacation_enabled

    if vacation_start_time is not None:
        data["vacation_start_time"] = vacation_start_time

    if vacation_end_time is not None:
        data["vacation_end_time"] = vacation_end_time

    if timezone is not None:
        data["timezone"] = timezone

    http.send_patch(url, headers, data)


def create_channel_stream_schedule_segment(
    token: str,
    client_id: str,
    broadcaster_id: str,
    start_time: datetime,
    timezone: str,
    is_recurring: bool,
    duration: int = 240,
    category_id: str | None = None,
    title: str | None = None,
) -> StreamSchedule:
    url = ENDPOINT_SEGMENTS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    payload = {
        "broadcaster_id": broadcaster_id,
        "start_time": start_time,
        "timezone": timezone,
        "is_recurring": is_recurring,
        "duration": duration,
    }

    if category_id is not None:
        payload["category_id"] = category_id

    if title is not None:
        payload["title"] = title

    schedule = http.send_post_get_result(url, headers, payload)[0]

    return StreamSchedule(
        [
            StreamScheduleSegment(
                segment["id"],
                datetime.strptime(segment["start_time"], date.RFC3339_FORMAT),
                datetime.strptime(segment["end_time"], date.RFC3339_FORMAT),
                segment["title"],
                datetime.strptime(segment["canceled_until"], date.RFC3339_FORMAT),
                Game(segment["category"]["id"], segment["category"]["name"]),
                segment["is_recurring"],
            )
            for segment in schedule["segments"]
        ],
        Channel(
            User(
                schedule["broadcaster_id"],
                schedule["broadcaster_login"],
                schedule["broadcaster_name"],
            )
        ),
        (
            datetime.strptime(schedule["vacation"]["start_time"], date.RFC3339_FORMAT),
            datetime.strptime(schedule["vacation"]["end_time"], date.RFC3339_FORMAT),
        ),
    )


def update_channel_stream_schedule_segment(
    token: str,
    client_id: str,
    broadcaster_id: str,
    stream_segment_id: str,
    start_time: datetime | None = None,
    duration: int | None = None,
    category_id: str | None = None,
    title: str | None = None,
    is_canceled: bool | None = None,
    timezone: str | None = None,
) -> StreamSchedule:
    url = ENDPOINT_SEGMENTS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {}
    data["broadcaster_id"] = broadcaster_id
    data["id"] = stream_segment_id

    if start_time is not None:
        data["start_time"] = start_time

    if duration is not None:
        data["duration"] = duration

    if category_id is not None:
        data["category_id"] = category_id

    if title is not None:
        data["title"] = title

    if is_canceled is not None:
        data["is_canceled"] = is_canceled

    if timezone is not None:
        data["timezone"] = timezone

    schedule = http.send_patch_get_result(url, headers, data)[0]

    return StreamSchedule(
        [
            StreamScheduleSegment(
                segment["id"],
                datetime.strptime(segment["start_time"], date.RFC3339_FORMAT),
                datetime.strptime(segment["end_time"], date.RFC3339_FORMAT),
                segment["title"],
                datetime.strptime(segment["canceled_until"], date.RFC3339_FORMAT),
                Game(segment["category"]["id"], segment["category"]["name"]),
                segment["is_recurring"],
            )
            for segment in schedule["segments"]
        ],
        Channel(
            User(
                schedule["broadcaster_id"],
                schedule["broadcaster_login"],
                schedule["broadcaster_name"],
            )
        ),
        (
            datetime.strptime(schedule["vacation"]["start_time"], date.RFC3339_FORMAT),
            datetime.strptime(schedule["vacation"]["end_time"], date.RFC3339_FORMAT),
        ),
    )


def delete_channel_stream_schedule_segment(
    token: str, client_id: str, broadcaster_id: str, stream_segment_id: str
) -> None:
    url = ENDPOINT_SEGMENTS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"broadcaster_id": broadcaster_id, "id": stream_segment_id}

    http.send_delete(url, headers, data)



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Channel


@dataclass
class Reward:
    """
    Represents a reward

    Attributes:
        channel (Channel): The channel that owns the reward
        reward_id (str): ID of the reward
        title (str | None): The title of the reward
        prompt (str | None): The prompt for the viewer when they are redeeming the reward
        cost (int | None): The cost of the reward
        image (dict | None): Image of the reward
        default_image (dict | None): Default images of the reward
        background_color (str | None): Background color of the reward
        is_enabled (bool | None): Is the reward currently enabled
        is_user_input_required (bool | None): Does the user need to enter information when redeeming the reward
        max_per_stream_setting (tuple[bool, int]): Settings about maximum uses per stream
        max_per_user_per_stream_setting (tuple[bool, int]): Settings about maximum uses per stream and user
        global_cooldown_setting (tuple[bool, int]): Settings about global cooldown
        is_paused (bool | None): Is the reward currently paused
        is_in_stock (bool | None): Is the reward currently in stock
        should_redemptions_skip_request_queue (bool | None): Should redemptions be set to FULFILLED status immediately when redeemed and skip the request queue instead of the normal UNFULFILLED status
        redemptions_redeemed_current_stream (int | None): The number of redemptions redeemed during the current live stream
        cooldown_expires_at (datetime | None): Timestamp of the cooldown expiration
    """

    channel: Channel
    reward_id: str
    title: str | None = None
    prompt: str | None = None
    cost: int | None = None
    image: dict | None = None
    default_image: dict | None = None
    background_color: str | None = None
    is_enabled: bool | None = None
    is_user_input_required: bool | None = None
    max_per_stream_setting: tuple[bool, int] | None = None
    max_per_user_per_stream_setting: tuple[bool, int] | None = None
    global_cooldown_setting: tuple[bool, int] | None = None
    is_paused: bool | None = None
    is_in_stock: bool | None = None
    should_redemptions_skip_request_queue: bool | None = None
    redemptions_redeemed_current_stream: int | None = None
    cooldown_expires_at: datetime | None = None



import math

import requests

from .. import errors

DEFAULT_TIMEOUT: int = 10


def send_post(url: str, headers: dict, payload: dict) -> None:
    response = requests.post(
        url, headers=headers, json=payload, timeout=DEFAULT_TIMEOUT
    )

    if not response.ok:
        raise errors.ClientError(response.json()["message"])


def send_post_get_result(url: str, headers: dict, payload: dict) -> list[dict]:
    response = requests.post(
        url, headers=headers, json=payload, timeout=DEFAULT_TIMEOUT
    )

    if not response.ok:
        raise errors.ClientError(response.json()["message"])

    return response.json()["data"]


def send_get(url: str, headers: dict, params: dict) -> list[dict]:
    response = requests.get(
        url, headers=headers, params=params, timeout=DEFAULT_TIMEOUT
    )

    if not response.ok:
        raise errors.ClientError(response.json()["message"])

    return response.json()["data"]


def send_get_with_pagination(
    url: str, headers: dict, params: dict, first: int, page_size: int
) -> list[dict]:
    after = ""
    results = []

    for call in range(math.ceil(first / page_size)):
        params["first"] = min(page_size, first - (page_size * call))

        if after != "":
            params["after"] = after

        response = requests.get(
            url, headers=headers, params=params, timeout=DEFAULT_TIMEOUT
        )

        if not response.ok:
            raise errors.ClientError(response.json()["message"])

        response = response.json()
        results.extend(response["data"])

        if "pagination" in response:
            after = response["pagination"]["cursor"]

    return results


def send_get_with_infinite_pagination(
    url: str, headers: dict, params: dict
) -> list[dict]:
    response = requests.get(
        url, headers=headers, params=params, timeout=DEFAULT_TIMEOUT
    )

    results = []

    while response.ok and "pagination" in response.json():
        results.extend(response.json()["data"])
        params["after"] = response.json()["pagination"]["cursor"]

        response = requests.get(
            url, headers=headers, params=params, timeout=DEFAULT_TIMEOUT
        )

    if not response.ok:
        raise errors.ClientError(response.json()["message"])

    return results


def send_get_text(url: str, params: dict) -> str:
    response = requests.get(url, params=params, timeout=DEFAULT_TIMEOUT)

    if not response.ok:
        raise errors.ClientError(response.json()["message"])

    return response.text


def send_put(url: str, headers: dict, data: dict) -> None:
    response = requests.put(url, headers=headers, data=data, timeout=DEFAULT_TIMEOUT)

    if not response.ok:
        raise errors.ClientError(response.json()["message"])


def send_put_get_result(url: str, headers: dict, data: dict) -> list[dict]:
    response = requests.put(url, headers=headers, data=data, timeout=DEFAULT_TIMEOUT)

    if not response.ok:
        raise errors.ClientError(response.json()["message"])

    return response.json()["data"]


def send_patch(url: str, headers: dict, data: dict) -> None:
    response = requests.patch(url, headers=headers, data=data, timeout=DEFAULT_TIMEOUT)

    if not response.ok:
        raise errors.ClientError(response.json()["message"])


def send_patch_get_result(url: str, headers: dict, data: dict) -> list[dict]:
    response = requests.patch(url, headers=headers, data=data, timeout=DEFAULT_TIMEOUT)

    if not response.ok:
        raise errors.ClientError(response.json()["message"])

    return response.json()["data"]


def send_delete(url: str, headers: dict, data: dict) -> None:
    response = requests.delete(url, headers=headers, data=data, timeout=DEFAULT_TIMEOUT)

    if not response.ok:
        raise errors.ClientError(response.json()["message"])


def send_delete_get_result(url: str, headers: dict, data: dict) -> list[dict]:
    response = requests.delete(url, headers=headers, data=data, timeout=DEFAULT_TIMEOUT)

    if not response.ok:
        raise errors.ClientError(response.json()["message"])

    return response.json()["data"]



from datetime import datetime

from .._utils import date, http
from ..dataclasses import Channel, Poll, PollChoice, User

ENDPOINT_POLLS = "https://api.twitch.tv/helix/polls"


def get_polls(
    token: str,
    client_id: str,
    broadcaster_id: str,
    poll_ids: list[str] | None = None,
    first: int = 20,
) -> list[Poll]:
    url = ENDPOINT_POLLS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}
    params["broadcaster_id"] = broadcaster_id

    if poll_ids is not None and len(poll_ids) > 0:
        params["id"] = poll_ids

    polls = http.send_get_with_pagination(url, headers, params, first, 20)

    return [
        Poll(
            poll["id"],
            Channel(
                User(
                    poll["broadcaster_id"],
                    poll["broadcaster_login"],
                    poll["broadcaster_name"],
                )
            ),
            poll["title"],
            [
                PollChoice(
                    choice["id"],
                    choice["title"],
                    choice["votes"],
                    choice["channel_points_votes"],
                )
                for choice in poll["choices"]
            ],
            poll["channel_points_voting_enabled"],
            poll["channel_points_per_vote"],
            poll["status"],
            poll["duration"],
            datetime.strptime(poll["started_at"], date.RFC3339_FORMAT),
            datetime.strptime(poll["ended_at"], date.RFC3339_FORMAT),
        )
        for poll in polls
    ]


def create_poll(
    token: str,
    client_id: str,
    broadcaster_id: str,
    title: str,
    choices: list[str],
    duration: int,
    channel_points_voting_enabled: bool = False,
    channel_points_per_vote: int = 0,
) -> Poll:
    url = ENDPOINT_POLLS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
        "Content-Type": "application/json",
    }

    choices_dicts = [{"title": choice} for choice in choices]

    payload = {
        "broadcaster_id": broadcaster_id,
        "title": title,
        "choices": choices_dicts,
        "duration": duration,
        "channel_points_voting_enabled": channel_points_voting_enabled,
        "channel_points_per_vote": channel_points_per_vote,
    }

    poll = http.send_post_get_result(url, headers, payload)[0]

    return Poll(
        poll["id"],
        Channel(
            User(
                poll["broadcaster_id"],
                poll["broadcaster_login"],
                poll["broadcaster_name"],
            )
        ),
        poll["title"],
        [
            PollChoice(
                choice["id"],
                choice["title"],
                choice["votes"],
                choice["channel_points_votes"],
            )
            for choice in poll["choices"]
        ],
        poll["channel_points_voting_enabled"],
        poll["channel_points_per_vote"],
        poll["status"],
        poll["duration"],
        datetime.strptime(poll["started_at"], date.RFC3339_FORMAT),
        datetime.strptime(poll["ended_at"], date.RFC3339_FORMAT),
    )


def end_poll(
    token: str, client_id: str, broadcaster_id: str, poll_id: str, status: str
) -> Poll:
    url = ENDPOINT_POLLS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"broadcaster_id": broadcaster_id, "id": poll_id, "status": status}

    poll = http.send_patch_get_result(url, headers, data)[0]

    return Poll(
        poll["id"],
        Channel(
            User(
                poll["broadcaster_id"],
                poll["broadcaster_login"],
                poll["broadcaster_name"],
            )
        ),
        poll["title"],
        [
            PollChoice(
                choice["id"],
                choice["title"],
                choice["votes"],
                choice["channel_points_votes"],
            )
            for choice in poll["choices"]
        ],
        poll["channel_points_voting_enabled"],
        poll["channel_points_per_vote"],
        poll["status"],
        poll["duration"],
        datetime.strptime(poll["started_at"], date.RFC3339_FORMAT),
        datetime.strptime(poll["ended_at"], date.RFC3339_FORMAT),
    )



from dataclasses import dataclass

from ..dataclasses import Transport


@dataclass
class ConduitShard:
    """
    Represents a conduit's shard

    Attributes:
        shard_id (str): Shard ID
        status (str): The shard status
            Possible values: enabled, webhook_callback_verification_pending, webhook_callback_verification_failed, notification_failures_exceeded, websocket_disconnected, websocket_failed_ping_pong, websocket_received_inbound_traffic, websocket_internal_error, websocket_network_timeout, websocket_network_error, websocket_failed_to_reconnect
        transport (Transport): The transport details used to send the notifications
    """

    shard_id: str
    status: str
    transport: Transport



from .._utils import http
from ..dataclasses import Tag


def get_all_stream_tags(
    token: str, client_id: str, first: int = 20, tag_id: list[str] | None = None
) -> list[Tag]:
    url = "https://api.twitch.tv/helix/tags/streams"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}

    if tag_id is not None and len(tag_id) > 0:
        params["tag_id"] = tag_id

    tags = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        Tag(
            tag["tag_id"],
            tag["is_auto"],
            tag["localization_names"],
            tag["localization_descriptions"],
        )
        for tag in tags
    ]


def get_stream_tags(token: str, client_id: str, broadcaster_id: str) -> list[Tag]:
    url = "https://api.twitch.tv/helix/streams/tags"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id}

    tags = http.send_get(url, headers, params)

    return [
        Tag(
            tag["tag_id"],
            tag["is_auto"],
            tag["localization_names"],
            tag["localization_descriptions"],
        )
        for tag in tags
    ]



from dataclasses import dataclass
from datetime import datetime


@dataclass
class GuestStarInvite:
    """
    Represents a Guest Star's invite

    Attributes:
        user_id (str): Twitch User ID corresponding to the invited guest
        invited_at (datetime): Timestamp when this user was invited to the session
        status (str): Status representing the invited user’s join state
            Possible values: INVITED, ACCEPTED, READY
        is_video_enabled (bool): Flag signaling that the invited user has chosen to disable their local video device
        is_audio_enabled (bool): Flag signaling that the invited user has chosen to disable their local audio device
        is_video_available (bool): Flag signaling that the invited user has a video device available for sharing
        is_audio_available (bool): Flag signaling that the invited user has an audio device available for sharing
    """

    user_id: str
    invited_at: datetime
    status: str
    is_video_enabled: bool
    is_audio_enabled: bool
    is_video_available: bool
    is_audio_available: bool



from .._utils import http
from ..dataclasses import Channel, Subscription, User


def get_broadcaster_subscriptions(
    token: str,
    client_id: str,
    broadcaster_id: str,
    user_id: list[str] | None = None,
    first: int = 20,
) -> list[Subscription]:
    url = "https://api.twitch.tv/helix/subscriptions"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}
    params["broadcaster_id"] = broadcaster_id

    if user_id is not None and len(user_id) > 0:
        params["user_id"] = user_id

    if first != 20:
        params["first"] = first

    subscriptions = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        Subscription(
            Channel(
                User(
                    subscription["broadcaster_id"],
                    subscription["broadcaster_login"],
                    subscription["broadcaster_name"],
                )
            ),
            User(
                subscription["gifter_id"],
                subscription["gifter_login"],
                subscription["gifter_name"],
            ),
            subscription["is_gift"],
            subscription["tier"],
            subscription["plan_name"],
            User(
                subscription["user_id"],
                subscription["user_login"],
                subscription["user_name"],
            ),
        )
        for subscription in subscriptions
    ]


def check_user_subscription(
    token: str, client_id: str, broadcaster_id: str, user_id: str
) -> Subscription:
    url = "https://api.twitch.tv/helix/subscriptions/user"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id, "user_id": user_id}

    subscription = http.send_get(url, headers, params)[0]

    return Subscription(
        Channel(
            User(
                subscription["broadcaster_id"],
                subscription["broadcaster_login"],
                subscription["broadcaster_name"],
            )
        ),
        User(
            subscription["gifter_id"],
            subscription["gifter_login"],
            subscription["gifter_name"],
        ),
        subscription["is_gift"],
        subscription["tier"],
    )



from dataclasses import dataclass


@dataclass
class GuestStarSettings:
    """
    Represents a Guest Star session's settings

    Attributes:
        is_moderator_send_live_enabled (bool): Flag determining if Guest Star moderators have access to control whether a guest is live once assigned to a slot
        slot_count (int): Number of slots the Guest Star call interface will allow the host to add to a call
        is_browser_source_audio_enabled (bool): Flag determining if Browser Sources subscribed to sessions on this channel should output audio
        group_layout (str): This setting determines how the guests within a session should be laid out within the browser source
            Possible values: TILED_LAYOUT, SCREENSHARE_LAYOUT
        browser_source_token (str): View only token to generate browser source URLs
    """

    is_moderator_send_live_enabled: bool
    slot_count: int
    is_browser_source_audio_enabled: bool
    group_layout: str
    browser_source_token: str



from dataclasses import dataclass


@dataclass
class BadgeVersion:
    """
    Represents a version of a chat badge

    Attributes:
        version_id (str): An ID that identifies this version of the badge
        image_url_1x (str): A URL to the small version (18px x 18px) of the badge
        image_url_2x (str): A URL to the medium version (36px x 36px) of the badge
        image_url_4x (str): A URL to the large version (72px x 72px) of the badge
        title (str): The title of the badge
        description (str): The description of the badge
        click_action (str): The action to take when clicking on the badge
        click_url (str): The URL to navigate to when clicking on the badge
    """

    version_id: str
    image_url_1x: str
    image_url_2x: str
    image_url_4x: str
    title: str
    description: str
    click_action: str
    click_url: str



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Channel, Game


@dataclass
class Stream:
    """
    Represents a stream

    Attributes:
        stream_id (str): An ID that identifies the stream
        channel (Channel): The channel that is broadcasting the stream
        game (Game): The category or game being played
        stream_type (str): The type of stream
        title (str): The stream’s title
        tags (list[str]): The tags applied to the stream
        viewer_count (int): The number of users watching the stream
        started_at (str): The UTC date and time (in RFC3339 format) of when the broadcast began
        language (str): The language that the stream uses
        thumbnail_url (str): A URL to an image of a frame from the last 5 minutes of the stream
        is_mature (bool): A Boolean value that indicates whether the stream is meant for mature audiences
    """

    stream_id: str
    channel: Channel
    game: Game
    stream_type: str
    title: str
    tags: list[str]
    viewer_count: int
    started_at: datetime
    language: str
    thumbnail_url: str
    is_mature: bool



from dataclasses import dataclass


@dataclass
class Tag:
    """
    Represents a stream tag

    Attributes:
        tag_id (str): An ID that identifies the tag
        is_auto (bool): A Boolean value that determines whether the tag is an automatic tag
            You cannot add or remove automatic tags
            The value is true if the tag is an automatic tag; otherwise, false
        localization_names (dict): The localized names of the tag
        localization_descriptions (dict): The localized descriptions of the tag
    """

    tag_id: str
    is_auto: bool
    localization_names: dict
    localization_descriptions: dict



from datetime import datetime

from .._utils import http
from ..dataclasses import BitsLeaderboardLeader, Cheermote, CheermoteTier, User


def get_bits_leaderboard(
    token: str,
    client_id: str,
    count: int = 10,
    period: str = "all",
    started_at: datetime | None = None,
    user_id: str | None = None,
) -> list[BitsLeaderboardLeader]:
    url = "https://api.twitch.tv/helix/bits/leaderboard"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"count": count, "period": period}

    if started_at is not None:
        params["started_at"] = started_at

    if user_id is not None:
        params["user_id"] = user_id

    leaderboard = http.send_get(url, headers, params)

    return [
        BitsLeaderboardLeader(
            User(leader["user_id"], leader["user_login"], leader["user_name"]),
            leader["rank"],
            leader["score"],
        )
        for leader in leaderboard
    ]


def get_cheermotes(
    token: str, client_id: str, broadcaster_id: str | None = None
) -> list[Cheermote]:
    url = "https://api.twitch.tv/helix/bits/cheermotes"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}

    if broadcaster_id is not None:
        params = {"broadcaster_id": broadcaster_id}

    cheermotes = http.send_get(url, headers, params)

    return [
        Cheermote(
            cheermote["prefix"],
            [
                CheermoteTier(
                    tier["min_bits"],
                    tier["tier_id"],
                    tier["color"],
                    tier["images"],
                    tier["can_cheer"],
                    tier["show_in_bits_card"],
                )
                for tier in cheermote["tier"]
            ],
            cheermote["type"],
            cheermote["order"],
            cheermote["last_updated"],
            cheermote["is_charitable"],
        )
        for cheermote in cheermotes
    ]



from .._utils import http
from ..dataclasses import Channel, Game, User


def search_categories(
    token: str, client_id: str, query: str, first: int = 20
) -> list[Game]:
    url = "https://api.twitch.tv/helix/search/categories"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"query": query}

    games = http.send_get_with_pagination(url, headers, params, first, 100)

    return [Game(game["id"], game["name"], game["box_art_url"]) for game in games]


def search_channels(
    token: str, client_id: str, query: str, first: int = 20, live_only: bool = False
) -> list[Channel]:
    url = "https://api.twitch.tv/helix/search/channels"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"query": query, "live_only": live_only}

    channels = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        Channel(
            User(channel["id"], channel["broadcaster_login"], channel["display_name"]),
            channel["broadcaster_language"],
            Game(channel["game_id"], channel["game_name"]),
            channel["title"],
            channel["tags"],
        )
        for channel in channels
    ]



from dataclasses import dataclass


@dataclass
class Extension:
    """
    Represents an extension

    Attributes:
        author_name (str): Name of the individual or organization that owns the Extension
        bits_enabled (bool): Whether the Extension has features that use Bits
        can_install (bool): Indicates if a user can install the Extension on their channel
            They may not be allowed if the Extension is currently in testing mode and the user is not on the allow list
        configuration_location (str): Whether the Extension configuration is hosted by the EBS or the Extensions Configuration Service
        description (str): The description of the Extension
        eula_tos_url (str): URL to the Extension’s Terms of Service
        has_chat_support (bool): Indicates if the Extension can communicate with the installed channel’s chat
        icon_url (str): The default icon to be displayed in the Extensions directory
        icon_urls (dict): The default icon in a variety of sizes
        extension_id (str): The autogenerated ID of the Extension
        name (str): The name of the Extension
        privacy_policy_url (str): URL to the Extension’s privacy policy
        request_identity_link (bool): Indicates if the Extension wants to explicitly ask viewers to link their Twitch identity
        screenshot_urls (list[str]): Screenshots to be shown in the Extensions marketplace
        state (str): The current state of the Extension
            Valid values are "InTest", "InReview", "Rejected", "Approved", "Released", "Deprecated", "PendingAction", "AssetsUploaded", "Deleted"
        subscriptions_support_level (str): Indicates if the Extension can determine a user’s subscription level on the channel the Extension is installed on
        summary (str): A brief description of the Extension
        support_email (str): The email users can use to receive Extension support
        version (str): The version of the Extension
        viewer_summary (str): A brief description displayed on the channel to explain how the Extension works
        views (dict): All configurations related to views such as: mobile, panel, video_overlay, and component
        allowlisted_config_urls (list[str]): Allow-listed configuration URLs for displaying the Extension
        allowlisted_panel_urls (list[str]): Allow-listed panel URLs for displaying the Extension
    """

    author_name: str
    bits_enabled: bool
    can_install: bool
    configuration_location: str
    description: str
    eula_tos_url: str
    has_chat_support: bool
    icon_url: str
    icon_urls: dict
    extension_id: str
    name: str
    privacy_policy_url: str
    request_identity_link: bool
    screenshot_urls: list[str]
    state: str
    subscriptions_support_level: str
    summary: str
    support_email: str
    version: str
    viewer_summary: str
    views: dict
    allowlisted_config_urls: list[str]
    allowlisted_panel_urls: list[str]



from dataclasses import dataclass


@dataclass
class CheermoteTier:
    """
    Represents a tier level that a Cheermote supports

    Attributes:
        min_bits (int): The minimum number of Bits that you must cheer at this tier level
            The maximum number of Bits that you can cheer at this level is determined by the required minimum Bits of the next tier level minus 1
            The minimum Bits value of the last tier is the maximum number of Bits you can cheer using this Cheermote
        tier_id (str): The tier level
            Possible values: 1, 100, 500, 1000, 5000, 10000, 100000
        color (str): The hex code of the color associated with this tier level
        images (dict): The animated and static image sets for the Cheermote
            The dictionary of images is organized by theme, format, and size
            The theme keys are dark and light
            Each theme is a dictionary of formats: animated and static
            Each format is a dictionary of sizes: 1, 1.5, 2, 3, and 4
            The value of each size contains the URL to the image
        can_cheer (bool): A Boolean value that determines whether users can cheer at this tier level
        show_in_bits_card (bool): A Boolean value that determines whether this tier level is shown in the Bits card
    """

    min_bits: int
    tier_id: str
    color: str
    images: dict
    can_cheer: bool
    show_in_bits_card: bool



from datetime import datetime

from .._utils import date, http
from ..dataclasses import (
    Guest,
    GuestStarInvite,
    GuestStarSession,
    GuestStarSettings,
    User,
)

ENDPOINT_SESSIONS = "https://api.twitch.tv/helix/guest_star/session"
ENDPOINT_INVITES = "https://api.twitch.tv/helix/guest_star/invites"
ENDPOINT_SLOTS = "https://api.twitch.tv/helix/guest_star/slot"


def get_channel_guest_star_settings(
    token: str, client_id: str, broadcaster_id: str, moderator_id: str
) -> GuestStarSettings:
    url = "https://api.twitch.tv/helix/guest_star/channel_settings"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id, "moderator_id": moderator_id}

    settings = http.send_get(url, headers, params)[0]

    return GuestStarSettings(
        settings["is_moderator_send_live_enabled"],
        settings["slot_count"],
        settings["is_browser_source_audio_enabled"],
        settings["group_layout"],
        settings["browser_source_token"],
    )


def update_channel_guest_star_settings(
    token: str,
    client_id: str,
    broadcaster_id: str,
    is_moderator_send_live_enabled: bool | None = None,
    slot_count: int | None = None,
    is_browser_source_audio_enabled: bool | None = None,
    group_layout: str | None = None,
    regenerate_browser_sources: bool | None = None,
) -> None:
    url = "https://api.twitch.tv/helix/guest_star/channel_settings"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {}
    data["broadcaster_id"] = broadcaster_id

    if is_moderator_send_live_enabled is not None:
        data["is_moderator_send_live_enabled"] = is_moderator_send_live_enabled

    if slot_count is not None:
        data["slot_count"] = slot_count

    if is_browser_source_audio_enabled is not None:
        data["is_browser_source_audio_enabled"] = is_browser_source_audio_enabled

    if group_layout is not None:
        data["group_layout"] = group_layout

    if regenerate_browser_sources is not None:
        data["regenerate_browser_sources"] = regenerate_browser_sources

    http.send_put(url, headers, data)


def get_guest_star_session(
    token: str, client_id: str, broadcaster_id: str, moderator_id: str
) -> GuestStarSession:
    url = ENDPOINT_SESSIONS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id, "moderator_id": moderator_id}

    session = http.send_get(url, headers, params)[0]

    return GuestStarSession(
        session["id"],
        [
            Guest(
                guest["slot_id"],
                guest["is_live"],
                User(guest["user_id"], guest["user_login"], guest["user_display_name"]),
                guest["volume"],
                datetime.strptime(guest["assigned_at"], date.RFC3339_FORMAT),
                guest["audio_settings"],
                guest["video_settings"],
            )
            for guest in session["guests"]
        ],
    )


def create_guest_star_session(
    token: str, client_id: str, broadcaster_id: str
) -> GuestStarSession:
    url = ENDPOINT_SESSIONS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    payload = {"broadcaster_id": broadcaster_id}

    session = http.send_post_get_result(url, headers, payload)[0]

    return GuestStarSession(
        session["id"],
        [
            Guest(
                guest["slot_id"],
                guest["is_live"],
                User(guest["user_id"], guest["user_login"], guest["user_display_name"]),
                guest["volume"],
                datetime.strptime(guest["assigned_at"], date.RFC3339_FORMAT),
                guest["audio_settings"],
                guest["video_settings"],
            )
            for guest in session["guests"]
        ],
    )


def end_guest_star_session(
    token: str, client_id: str, broadcaster_id: str, session_id: str
) -> GuestStarSession:
    url = ENDPOINT_SESSIONS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"broadcaster_id": broadcaster_id, "session_id": session_id}

    session = http.send_delete_get_result(url, headers, data)[0]

    return GuestStarSession(
        session["id"],
        [
            Guest(
                guest["slot_id"],
                guest["is_live"],
                User(guest["user_id"], guest["user_login"], guest["user_display_name"]),
                guest["volume"],
                datetime.strptime(guest["assigned_at"], date.RFC3339_FORMAT),
                guest["audio_settings"],
                guest["video_settings"],
            )
            for guest in session["guests"]
        ],
    )


def get_guest_star_invites(
    token: str, client_id: str, broadcaster_id: str, moderator_id: str, session_id: str
) -> list[GuestStarInvite]:
    url = ENDPOINT_INVITES
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {
        "broadcaster_id": broadcaster_id,
        "moderator_id": moderator_id,
        "session_id": session_id,
    }

    invites = http.send_get(url, headers, params)

    return [
        GuestStarInvite(
            invite["user_id"],
            datetime.strptime(invite["invited_at"], date.RFC3339_FORMAT),
            invite["status"],
            invite["is_video_enabled"],
            invite["is_audio_enabled"],
            invite["is_video_available"],
            invite["is_audio_available"],
        )
        for invite in invites
    ]


def send_guest_star_invite(
    token: str,
    client_id: str,
    broadcaster_id: str,
    moderator_id: str,
    session_id: str,
    guest_id: str,
) -> None:
    url = ENDPOINT_INVITES
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    payload = {
        "broadcaster_id": broadcaster_id,
        "moderator_id": moderator_id,
        "session_id": session_id,
        "guest_id": guest_id,
    }

    http.send_post(url, headers, payload)


def delete_guest_star_invite(
    token: str,
    client_id: str,
    broadcaster_id: str,
    moderator_id: str,
    session_id: str,
    guest_id: str,
) -> None:
    url = ENDPOINT_INVITES
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {
        "broadcaster_id": broadcaster_id,
        "moderator_id": moderator_id,
        "session_id": session_id,
        "guest_id": guest_id,
    }

    http.send_delete(url, headers, data)


def assign_guest_star_slot(
    token: str,
    client_id: str,
    broadcaster_id: str,
    moderator_id: str,
    session_id: str,
    guest_id: str,
    slot_id: str,
) -> None:
    url = ENDPOINT_SLOTS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    payload = {
        "broadcaster_id": broadcaster_id,
        "moderator_id": moderator_id,
        "session_id": session_id,
        "guest_id": guest_id,
        "slot_id": slot_id,
    }

    http.send_post(url, headers, payload)


def update_guest_star_slot(
    token: str,
    client_id: str,
    broadcaster_id: str,
    moderator_id: str,
    session_id: str,
    source_slot_id: str,
    destination_slot_id: str | None = None,
) -> None:
    url = ENDPOINT_SLOTS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {
        "broadcaster_id": broadcaster_id,
        "moderator_id": moderator_id,
        "session_id": session_id,
        "source_slot_id": source_slot_id,
    }

    if destination_slot_id is not None:
        data["destination_slot_id"] = destination_slot_id

    http.send_patch(url, headers, data)


def delete_guest_star_slot(
    token: str,
    client_id: str,
    broadcaster_id: str,
    moderator_id: str,
    session_id: str,
    guest_id: str,
    slot_id: str,
    should_reinvite_guest: str | None = None,
) -> None:
    url = ENDPOINT_SLOTS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {
        "broadcaster_id": broadcaster_id,
        "moderator_id": moderator_id,
        "session_id": session_id,
        "guest_id": guest_id,
        "slot_id": slot_id,
    }

    if should_reinvite_guest is not None:
        data["should_reinvite_guest"] = should_reinvite_guest

    http.send_delete(url, headers, data)


def update_guest_star_slot_settings(
    token: str,
    client_id: str,
    broadcaster_id: str,
    moderator_id: str,
    session_id: str,
    slot_id: str,
    is_audio_enabled: bool | None = None,
    is_video_enabled: bool | None = None,
    is_live: bool | None = None,
    volume: int | None = None,
) -> None:
    url = "https://api.twitch.tv/helix/guest_star/slot_settings"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {}
    data["broadcaster_id"] = broadcaster_id
    data["moderator_id"] = moderator_id
    data["session_id"] = session_id
    data["slot_id"] = slot_id

    if is_audio_enabled is not None:
        data["is_audio_enabled"] = is_audio_enabled

    if is_video_enabled is not None:
        data["is_video_enabled"] = is_video_enabled

    if is_live is not None:
        data["is_live"] = is_live

    if volume is not None:
        data["volume"] = volume

    http.send_patch(url, headers, data)



from dataclasses import dataclass
from datetime import datetime


@dataclass
class User:
    """
    Represents an user

    Attributes:
        user_id (int): User's ID
        login (str): User's login
        display_name (str): User's name
        user_type (str | None): User type
        broadcaster_type (str | None): User's range
        description (str | None): User's description
        profile_image_url (str | None): URL of the user's profile image
        offline_image_url (str | None): URL of the image that is displayed when the user is not on stream
        view_count (int | None): Number of user viewers
        email (str | None): The user’s verified email address
        created_at (datetime | None): The UTC date and time that the user’s account was created
            The timestamp is in RFC3339 format
    """

    user_id: int
    login: str
    display_name: str
    user_type: str | None = None
    broadcaster_type: str | None = None
    description: str | None = None
    profile_image_url: str | None = None
    offline_image_url: str | None = None
    view_count: int | None = None
    email: str | None = None
    created_at: datetime | None = None



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Channel, PredictionOutcome


@dataclass
class Prediction:
    """
    Represents a prediction

    Args:
        prediction_id (str): ID of the Prediction
        channel (Channel): The channel where the prediction was created
        title (str): Title for the Prediction
        winning_outcome_id (str): ID of the winning outcome
            If the status is ACTIVE, this is set to null
        outcomes (list[PredictionOutcome]): Possible outcomes for the Prediction
        prediction_window (int): Total duration for the Prediction (in seconds)
        status (str): Status of the Prediction
            Valid values are: RESOLVED, ACTIVE, CANCELED, LOCKED
        created_at (datetime): UTC timestamp for the Prediction’s start time
        ended_at (datetime): UTC timestamp for when the Prediction ended
            If the status is ACTIVE, this is set to null
        locked_at (datetime): UTC timestamp for when the Prediction was locked
            If the status is not LOCKED, this is set to null
    """

    prediction_id: str
    channel: Channel
    title: str
    winning_outcome_id: str
    outcomes: list[PredictionOutcome]
    prediction_window: int
    status: str
    created_at: datetime
    ended_at: datetime
    locked_at: datetime



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import User


@dataclass
class ShieldModeStatus:
    """
    Represents the Shield Mode status

    Attributes:
        is_active (bool): A Boolean value that determines whether Shield Mode is active
        moderator (User): The moderator that last activated Shield Mode.
        last_activated_at (datetime): The UTC timestamp (in RFC3339 format) of when Shield Mode was last activated
    """

    is_active: bool
    moderator: User
    last_activated_at: datetime



from dataclasses import dataclass

from ..dataclasses import Game, User


@dataclass
class Channel:
    """
    Represents a channel

    Attributes:
        user (User): The user associated with the channel
        broadcaster_language (str | None): The broadcaster’s preferred language
        game (Game | None): The game that the broadcaster is playing or last played
        title (str | None): The title of the stream that the broadcaster is currently streaming or last streamed
        tags (list[str] | None): The tags applied to the channel
        delay (int | None): The value of the broadcaster’s stream delay setting, in seconds
        content_classification_labels (list[str] | None): The CCLs applied to the channel
        is_branded_content (bool | None): Boolean flag indicating if the channel has branded content
    """

    user: User
    broadcaster_language: str | None = None
    game: Game | None = None
    title: str | None = None
    tags: list[str] | None = None
    delay: int | None = None
    content_classification_labels: list[str] | None = None
    is_branded_content: bool | None = None



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Channel, StreamScheduleSegment


@dataclass
class StreamSchedule:
    """
    Represents a stream schedule

    Attributes:
        segments (list[StreamScheduleSegment]): Scheduled broadcasts for this stream schedule
        channel (Channel): The channel that owns the broadcast schedule
        vacation (tuple[datetime, datetime]): If Vacation Mode is enabled, this includes start and end dates for the vacation
            If Vacation Mode is disabled, value is set to null
    """

    segments: list[StreamScheduleSegment]
    channel: Channel
    vacation: tuple[datetime, datetime]



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Channel, User


@dataclass
class UnbanRequest:
    """
    Represents an unban request

    Attributes:
        request_id (str): Unban request ID
        channel (Channel): Channel that is receiving the unban request
        moderator (User): Moderator who approved/denied the request
        user (User): User who is asking for an unban
        text (str): Text of the request from the requesting user
        status (str): Status of the request
            Possible values: pending, approved, denied, acknowledged, canceled
        created_at (datetime): Timestamp of when the unban request was created
        resolved_at (datetime): Timestamp of when moderator/broadcaster approved or denied the request
        resolution_text (str): Text input by the resolver (moderator) of the unban request
    """

    request_id: str
    channel: Channel
    moderator: User
    user: User
    text: str
    status: str
    created_at: datetime
    resolved_at: datetime
    resolution_text: str



from datetime import datetime

from .._utils import date, http
from ..dataclasses import User

ENDPOINT_BLOCKS = "https://api.twitch.tv/helix/users/blocks"


def get_users(
    token: str,
    client_id: str,
    user_ids: list[str] | None = None,
    login: list[str] | None = None,
) -> list[User]:
    url = "https://api.twitch.tv/helix/users"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}

    if user_ids is not None and len(user_ids) > 0:
        params["id"] = user_ids

    if login is not None and len(login) > 0:
        params["login"] = [user_login.replace("@", "").lower() for user_login in login]

    users = http.send_get(url, headers, params)

    return [
        User(
            user["id"],
            user["login"],
            user["display_name"],
            user["type"],
            user["broadcaster_type"],
            user["description"],
            user["profile_image_url"],
            user["offline_image_url"],
            user["view_count"],
            user["email"] if "email" in user else None,
            datetime.strptime(user["created_at"], date.RFC3339_FORMAT),
        )
        for user in users
    ]


def update_user(token: str, client_id: str, description: str | None = None) -> User:
    url = "https://api.twitch.tv/helix/users"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {}

    if description != "":
        data = {"description": description}

    user = http.send_put_get_result(url, headers, data)[0]

    return User(
        user["id"],
        user["login"],
        user["display_name"],
        user["type"],
        user["broadcaster_type"],
        user["description"],
        user["profile_image_url"],
        user["offline_image_url"],
        user["view_count"],
        user["email"],
        datetime.strptime(user["created_at"], date.RFC3339_FORMAT),
    )


def get_user_block_list(
    token: str, client_id: str, broadcaster_id: str, first: int = 20
) -> list[User]:
    url = ENDPOINT_BLOCKS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}
    params["broadcaster_id"] = broadcaster_id

    if first != 20:
        params["first"] = first

    users = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        User(user["user_id"], user["user_login"], user["display_name"])
        for user in users
    ]


def block_user(
    token: str,
    client_id: str,
    target_user_id: str,
    source_context: str | None = None,
    reason: str | None = None,
) -> None:
    url = ENDPOINT_BLOCKS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"target_user_id": target_user_id}

    if source_context is not None:
        data["source_context"] = source_context

    if reason is not None:
        data["reason"] = reason

    http.send_put(url, headers, data)


def unblock_user(token: str, client_id: str, target_user_id: str) -> None:
    url = ENDPOINT_BLOCKS
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    data = {"target_user_id": target_user_id}

    http.send_delete(url, headers, data)


def get_user_extensions(token: str, client_id: str) -> list[dict]:
    url = "https://api.twitch.tv/helix/users/extensions/list"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }

    return http.send_get(url, headers, {})


def get_user_active_extensions(
    token: str, client_id: str, user_id: str | None = None
) -> list[dict]:
    url = "https://api.twitch.tv/helix/users/extensions"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}

    if user_id != "":
        params = {"user_id": user_id}

    return http.send_get(url, headers, params)


def update_user_extensions(token: str, client_id: str, data: dict) -> list[dict]:
    url = "https://api.twitch.tv/helix/users/extensions"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
        "Content-Type": "application/json",
    }
    data = {"data": data}

    return http.send_put_get_result(url, headers, {})



from datetime import datetime

from .._utils import date, http
from ..dataclasses import Channel, Game, Stream, StreamMarker, User


def get_stream_key(token: str, client_id: str, broadcaster_id: str) -> str:
    url = "https://api.twitch.tv/helix/streams/key"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"broadcaster_id": broadcaster_id}

    return http.send_get(url, headers, params)[0]["stream_key"]


def get_streams(
    token: str,
    client_id: str,
    user_id: list[str] | None = None,
    user_login: list[str] | None = None,
    game_id: list[str] | None = None,
    stream_type: str = "all",
    language: list[str] | None = None,
    first: int = 20,
) -> list[Stream]:
    url = "https://api.twitch.tv/helix/streams"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}
    params["stream_type"] = stream_type

    if user_id is not None:
        params["user_id"] = user_id

    if user_login is not None:
        params["user_login"] = user_login

    if game_id is not None:
        params["game_id"] = game_id

    if language is not None:
        params["language"] = language

    streams = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        Stream(
            stream["id"],
            Channel(User(stream["user_id"], stream["user_login"], stream["user_name"])),
            Game(stream["game_id"], stream["game_name"]),
            stream["type"],
            stream["title"],
            stream["tags"],
            stream["viewer_count"],
            datetime.strptime(stream["started_at"], date.RFC3339_FORMAT),
            stream["language"],
            stream["thumbnail_url"],
            stream["is_mature"],
        )
        for stream in streams
    ]


def get_followed_streams(
    token: str, client_id: str, user_id: str, first: int = 100
) -> list[Stream]:
    url = "https://api.twitch.tv/helix/streams/followed"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {"user_id": user_id}

    streams = http.send_get_with_pagination(url, headers, params, first, 100)

    return [
        Stream(
            stream["id"],
            Channel(User(stream["user_id"], stream["user_login"], stream["user_name"])),
            Game(stream["game_id"], stream["game_name"]),
            stream["type"],
            stream["title"],
            stream["tags"],
            stream["viewer_count"],
            datetime.strptime(stream["started_at"], date.RFC3339_FORMAT),
            stream["language"],
            stream["thumbnail_url"],
            stream["is_mature"],
        )
        for stream in streams
    ]


def create_stream_marker(
    token: str, client_id: str, user_id: str, description: str | None = None
) -> StreamMarker:
    url = "https://api.twitch.tv/helix/streams/markers"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
        "Content-Type": "application/json",
    }
    payload = {"user_id": user_id}

    if description is not None:
        payload["description"] = description

    marker = http.send_post_get_result(url, headers, payload)[0]

    return StreamMarker(
        marker["id"],
        datetime.strptime(marker["created_at"], date.RFC3339_FORMAT),
        marker["position_seconds"],
        marker["description"],
    )


def get_stream_markers(
    token: str,
    client_id: str,
    user_id: str | None = None,
    video_id: str | None = None,
    first: int = 20,
) -> list[dict]:
    url = "https://api.twitch.tv/helix/streams/markers"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": client_id,
    }
    params = {}

    if user_id is not None:
        params["user_id"] = user_id

    if video_id is not None:
        params["video_id"] = video_id

    return http.send_get_with_pagination(url, headers, params, first, 100)



from dataclasses import dataclass

from ..dataclasses import Channel, User


@dataclass
class Clip:
    """
    Represents a clip

    Attributes:
        clip_id (str): An ID that uniquely identifies the clip
        url (str): A URL to the clip
        embed_url (str): A URL that you can use in an iframe to embed the clip
        channel (Channel): The channel that the video was clipped from
        creator (User): The user that created the clip
        video_id (str): An ID that identifies the video that the clip came from
        game_id (str): The ID of the game that was being played when the clip was created
        language (str): The ISO 639-1 two-letter language code that the broadcaster broadcasts in
        title (str): The title of the clip
        view_count (int): The number of times the clip has been viewed
        created_at (str): The date and time of when the clip was created
        thumbnail_url (str): A URL to a thumbnail image of the clip
        duration (float): The length of the clip, in seconds
            Precision is 0.1
        vod_offset (int): The zero-based offset, in seconds, to where the clip starts in the video (VOD)
        is_featured (bool): A Boolean value that indicates if the clip is featured or not
    """

    clip_id: str
    url: str
    embed_url: str
    channel: Channel
    creator: User
    video_id: str
    game_id: str
    language: str
    title: str
    view_count: int
    created_at: str
    thumbnail_url: str
    duration: float
    vod_offset: int
    is_featured: bool



from dataclasses import dataclass


@dataclass
class CharityCampaignAmount:
    """
    Represents the amount of a donation in a charity campaign

    Attributes:
        value (int): The monetary amount
            The amount is specified in the currency’s minor unit
        decimal_places (int): The number of decimal places used by the currency
        currency (str): The ISO-4217 three-letter currency code that identifies the type of currency in value
    """

    value: int
    decimal_places: int
    currency: str



from dataclasses import dataclass


@dataclass
class AutoModSettings:
    """
    Represents the AutoMod settings

    Attributes:
        broadcaster_id (str): The broadcaster’s ID
        moderator_id (str): The moderator’s ID
        overall_level (int): The default AutoMod level for the broadcaster
        disability (int): The Automod level for discrimination against disability
        aggression (int): The Automod level for hostility involving aggression
        sexuality_sex_or_gender (int): The AutoMod level for discrimination based on sexuality, sex, or gender
        misogyny (int): The Automod level for discrimination against women
        bullying (int): The Automod level for hostility involving name calling or insults
        swearing (int): The Automod level for profanity
        race_ethnicity_or_religion (int): The Automod level for racial discrimination
        sex_based_terms (int): The Automod level for sexual content
    """

    broadcaster_id: str
    moderator_id: str
    overall_level: int
    disability: int
    aggression: int
    sexuality_sex_or_gender: int
    misogyny: int
    bullying: int
    swearing: int
    race_ethnicity_or_religion: int
    sex_based_terms: int



from dataclasses import dataclass
from datetime import datetime


@dataclass
class AdSchedule:
    """
    Represents an ad schecule

    Attributes:
        snooze_count (int): The number of snoozes available for the broadcaster
        snooze_refresh_at (datetime): The UTC timestamp when the broadcaster will gain an additional snooze, in RFC3339 format
        next_ad_at (datetime): The UTC timestamp of the broadcaster’s next scheduled ad, in RFC3339 format
        duration (int): The length in seconds of the scheduled upcoming ad break
        last_ad_at (datetime): The UTC timestamp of the broadcaster’s last ad-break, in RFC3339 format
        preroll_free_time (int): The amount of pre-roll free time remaining for the channel in seconds
    """

    snooze_count: int
    snooze_refresh_at: datetime
    next_ad_at: datetime
    duration: int | None = None
    last_ad_at: datetime | None = None
    preroll_free_time: int | None = None



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Channel, Product, User


@dataclass
class ExtensionTransaction:
    """
    Represents an extension's transaction

    Attributes:
        transaction_id (str): An ID that identifies the transaction
        timestamp (datetime): The UTC date and time (in RFC3339 format) of the transaction
        channel (Channel): The channel where the transaction occurred
        user (User): The user that purchased the digital product
        product_type (str): The type of transaction
            Possible values: BITS_IN_EXTENSION
        product (Product): Contains details about the digital product
    """

    transaction_id: str
    timestamp: datetime
    channel: Channel
    user: User
    product_type: str
    product: Product



from dataclasses import dataclass


@dataclass
class PollChoice:
    """
    Represents a choice in a poll

    Attributes:
        choice_id (str): An ID that identifies this choice
        title (str): The choice's title
            The title may contain a maximum of 25 characters
        votes (int): The total number of votes cast for this choice
        channel_points_votes (int): The number of votes cast using Channel Points
    """

    choice_id: str
    title: str
    votes: int
    channel_points_votes: int



from dataclasses import dataclass
from datetime import datetime

from ..dataclasses import Channel


@dataclass
class CreatorGoal:
    """
    Represents a creator's goal

    Attributes:
        goal_id (str): An ID that identifies this goal
        channel (Channel): The channel where the goal was created
        type (str): The type of goal
            Possible values: follower, subscription, subscription_count, new_subscription, new_subscription_count
        description (str): A description of the goal
        current_amount (int): The goal’s current value
        target_amount (int): The goal’s target value
        created_at (datetime): The UTC date and time (in RFC3339 format) that the broadcaster created the goal
    """

    goal_id: str
    channel: Channel
    type: str
    description: str
    current_amount: int
    target_amount: int
    created_at: datetime
